from abc import ABCMeta
import itertools
import numpy as np
from pprint import pformat
from typing import Dict, Iterable, List, Tuple

from pyRDDLGym.core.debug.exception import (
    print_stack_trace_root as PST,
    RDDLInvalidExpressionError,
    RDDLInvalidNumberOfArgumentsError,
    RDDLInvalidObjectError,
    RDDLMissingCPFDefinitionError,
    RDDLRepeatedVariableError,
    RDDLTypeError,
    RDDLUndefinedCPFError,
    RDDLUndefinedVariableError,
    RDDLValueOutOfRangeError
)
from pyRDDLGym.core.parser.expr import Expression, Value


class RDDLPlanningModel(metaclass=ABCMeta):
    '''The base class representing all RDDL domains + instances.
    '''
    
    # grounded variable is var___obj1__obj2...
    FLUENT_SEP = '___'
    OBJECT_SEP = '__'
    NEXT_STATE_SYM = '\''

    PRIMITIVE_TYPES = {
        'int': int,
        'real': float,
        'bool': bool
    }
    
    def __init__(self):
        
        # base
        self._AST = None
        
        # objects
        self._type_to_objects = None
        self._object_to_type = None
        self._object_to_index = None
        self._enum_types = None
        
        # variable info
        self._variable_types = None
        self._variable_ranges = None
        self._variable_params = None
        self._variable_defaults = None
        self._variable_groundings = None     
        
        self._non_fluents = None        
        self._state_fluents = None
        self._state_ranges = None
        self._prev_state = None
        self._next_state = None    
        self._action_fluents = None
        self._action_ranges = None        
        self._derived_fluents = None
        self._interm_fluents = None        
        self._observ_fluents = None
        self._observ_ranges = None
        
        # cpf info
        self._cpfs = None
        self._level_to_cpfs = None
        self._cpf_to_level = None
        self._reward = None
        
        # constraint info
        self._terminations = None
        self._preconditions = None
        self._invariants = None
        
        # instance info
        self._discount = None
        self._horizon = None
        self._max_allowed_actions = None        
        
    # ===========================================================================
    # base properties
    # ===========================================================================
    
    @property
    def ast(self):
        return self._AST

    @ast.setter
    def ast(self, value):
        self._AST = value
    
    @property
    def domain_name(self):
        if self.ast is None:
            return None
        else:
            return self.ast.domain.name
    
    @property
    def instance_name(self):
        if self.ast is None:
            return None
        else:
            return self.ast.instance.name
    
    # ===========================================================================
    # objects
    # ===========================================================================
    
    @property
    def type_to_objects(self):
        return self._type_to_objects

    @type_to_objects.setter
    def type_to_objects(self, value):
        self._type_to_objects = value
    
    @property
    def object_to_type(self):
        return self._object_to_type

    @object_to_type.setter
    def object_to_type(self, value):
        self._object_to_type = value
    
    @property
    def object_to_index(self):
        return self._object_to_index

    @object_to_index.setter
    def object_to_index(self, val):
        self._object_to_index = val
    
    @property
    def enum_types(self):
        return self._enum_types

    @enum_types.setter
    def enum_types(self, value):
        self._enum_types = value
    
    # ===========================================================================
    # variables
    # ===========================================================================
    
    @property
    def variable_types(self):
        return self._variable_types

    @variable_types.setter
    def variable_types(self, val):
        self._variable_types = val
    
    @property
    def variable_ranges(self):
        return self._variable_ranges

    @variable_ranges.setter
    def variable_ranges(self, val):
        self._variable_ranges = val
    
    @property
    def variable_params(self):
        return self._variable_params

    @variable_params.setter
    def variable_params(self, val):
        self._variable_params = val
    
    @property
    def variable_defaults(self):
        return self._variable_defaults
    
    @variable_defaults.setter
    def variable_defaults(self, val):
        self._variable_defaults = val
        
    @property
    def variable_groundings(self):
        return self._variable_groundings

    @variable_groundings.setter
    def variable_groundings(self, val):
        self._variable_groundings = val
    
    @property
    def non_fluents(self):
        return self._non_fluents

    @non_fluents.setter
    def non_fluents(self, val):
        self._non_fluents = val

    @property
    def state_fluents(self):
        return self._state_fluents

    @state_fluents.setter
    def state_fluents(self, val):
        self._state_fluents = val
    
    @property
    def state_ranges(self):
        return self._state_ranges

    @state_ranges.setter
    def state_ranges(self, value):
        self._state_ranges = value

    @property
    def next_state(self):
        return self._next_state

    @next_state.setter
    def next_state(self, val):
        self._next_state = val

    @property
    def prev_state(self):
        return self._prev_state

    @prev_state.setter
    def prev_state(self, val):
        self._prev_state = val

    @property
    def action_fluents(self):
        return self._action_fluents

    @action_fluents.setter
    def action_fluents(self, val):
        self._action_fluents = val

    @property
    def action_ranges(self):
        return self._action_ranges

    @action_ranges.setter
    def action_ranges(self, value):
        self._action_ranges = value
        
    @property
    def derived_fluents(self):
        return self._derived_fluents

    @derived_fluents.setter
    def derived_fluents(self, val):
        self._derived_fluents = val

    @property
    def interm_fluents(self):
        return self._interm_fluents

    @interm_fluents.setter
    def interm_fluents(self, val):
        self._interm_fluents = val
    
    @property
    def observ_fluents(self):
        return self._observ_fluents

    @observ_fluents.setter
    def observ_fluents(self, value):
        self._observ_fluents = value

    @property
    def observ_ranges(self):
        return self._observ_ranges

    @observ_ranges.setter
    def observ_ranges(self, value):
        self._observ_ranges = value
    
    # ===========================================================================
    # CPFs
    # ===========================================================================
    
    @property
    def cpfs(self):
        return self._cpfs

    @cpfs.setter
    def cpfs(self, val):
        self._cpfs = val

    @property
    def level_to_cpfs(self):
        return self._level_to_cpfs

    @level_to_cpfs.setter
    def level_to_cpfs(self, val):
        self._level_to_cpfs = val

    @property
    def cpf_to_level(self):
        return self._cpf_to_level

    @cpf_to_level.setter
    def cpf_to_level(self, val):
        self._cpf_to_level = val
    
    @property
    def reward(self):
        return self._reward

    @reward.setter
    def reward(self, val):
        self._reward = val
    
    # ===========================================================================
    # constraints
    # ===========================================================================
    
    @property
    def terminations(self):
        return self._terminations

    @terminations.setter
    def terminations(self, value):
        self._terminations = value

    @property
    def preconditions(self):
        return self._preconditions

    @preconditions.setter
    def preconditions(self, val):
        self._preconditions = val

    @property
    def invariants(self):
        return self._invariants

    @invariants.setter
    def invariants(self, val):
        self._invariants = val
    
    # ===========================================================================
    # init-block
    # ===========================================================================
    
    @property
    def discount(self):
        return self._discount

    @discount.setter
    def discount(self, val):
        self._discount = val

    @property
    def horizon(self):
        return self._horizon

    @horizon.setter
    def horizon(self, val):
        self._horizon = val

    @property
    def max_allowed_actions(self):
        return self._max_allowed_actions

    @max_allowed_actions.setter
    def max_allowed_actions(self, val):
        self._max_allowed_actions = val
    
    # ===========================================================================
    # class methods for general RDDL syntax rules
    # ===========================================================================
    
    @staticmethod
    def is_free_object(name: str) -> bool:
        '''Determines whether the name is a free object (e.g., ?x).
        '''
        return name[0] == '?'
    
    @staticmethod
    def strip_literal(name: str) -> str:
        '''Returns the canonical name of an enum literal 
        (e.g., given @x returns x). All other strings are returned unmodified.
        '''
        if name[0] == '@':
            name = name[1:]
        return name
    
    @staticmethod
    def strip_literals(names: Iterable[str]) -> List[str]:
        '''Returns the canonical names of given enum literals 
        (e.g., given @x returns x). All other strings are returned unmodified.
        '''
        return list(map(RDDLPlanningModel.strip_literal, names))
        
    @staticmethod
    def ground_var(name: str, objects: Iterable[str]) -> str:
        '''Given a variable name and list of objects as arguments, produces the
        grounded representation <variable>___<obj1>__<obj2>__...
        '''
        PRIME = RDDLPlanningModel.NEXT_STATE_SYM
        is_primed = name.endswith(PRIME)
        var = name
        if is_primed:
            var = var[:-len(PRIME)]
        if objects is not None and objects:
            objects = RDDLPlanningModel.OBJECT_SEP.join(objects)
            var += RDDLPlanningModel.FLUENT_SEP + objects
        if is_primed:
            var += PRIME
        return var
    
    @staticmethod
    def parse_grounded(expr: str) -> Tuple[str, List[str]]:
        '''Parses a variable of the form <name>___<type1>__<type2>...
        into a tuple (<name>, [<type1>, <type2>, ...]).
        '''
        PRIME = RDDLPlanningModel.NEXT_STATE_SYM
        is_primed = expr.endswith(PRIME)
        if is_primed:
            expr = expr[:-len(PRIME)]
        var, *objects = expr.split(RDDLPlanningModel.FLUENT_SEP)
        if objects:
            if len(objects) != 1:
                raise RDDLInvalidObjectError(
                    f'Variable {expr} contains multiple fluent separators.')
            objects = objects[0].split(RDDLPlanningModel.OBJECT_SEP)
        if is_primed:
            var += PRIME
        return var, objects
    
    # ===========================================================================
    # utility methods
    # ===========================================================================
    
    def is_type(self, value: str) -> bool:
        '''Returns whether the given value is a valid type.
        '''
        return value in self.type_to_objects or value in self.enum_types
        
    def is_object(self, name: str, msg='') -> bool:
        '''Returns whether the given name is a valid object.
        
        Raises an exception if an object is identified as an enum literal (@x)
        but is not defined, or if it is an object that shares
        the same name as a pvariable with no parameters.
        '''        
        # must be an object
        if name[0] == '@':
            name = name[1:]
            if name not in self.object_to_type:
                raise RDDLInvalidObjectError(
                    f'Object <@{name}> is not defined, '
                    f'must be one of {set(self.object_to_type.keys())}. {msg}')
            return True
        
        # this could either be an object or variable
        # object if a variable does not exist with the same name
        #        or if a variable has the same name but free parameters
        # ambiguous if a variable has the same name but no free parameters
        if name in self.object_to_type:
            params = self.variable_params.get(name, None)
            if params is not None and not params:
                raise RDDLInvalidObjectError(
                    f'Ambiguous reference to object or '
                    f'parameter-free variable with identical name <{name}>. {msg}')
            return True
        
        # this must be something else...
        return False
    
    def is_literal(self, name: str) -> bool:
        '''Returns whether the given name is a valid enumerated (domain) object.
        '''
        if RDDLPlanningModel.is_free_object(name):
            return False
        name = RDDLPlanningModel.strip_literal(name)
        ptype = self.object_to_type.get(name, None)
        if ptype is None:
            return False
        return ptype in self.enum_types
        
    def object_indices(self, objects: Iterable[str], msg: str='') -> Tuple[int, ...]:
        '''Returns the canonical indices of a sequence of objects based on the
        orders they are defined in the instance.
        
        Raises an exception if an object is not defined in the domain.
        '''
        object_to_index = self.object_to_index
        try:
            return tuple(object_to_index[obj] for obj in objects)
        except:
            for obj in objects:
                if obj not in object_to_index:
                    raise RDDLInvalidObjectError(
                        f'Object <{obj}> is not valid, '
                        f'must be one of {set(object_to_index.keys())}. {msg}')
    
    def object_counts(self, types: Iterable[str], msg: str='') -> Tuple[int, ...]:
        '''Returns a tuple containing the number of objects of each type.
        Raises an exception if a type is not defined in the domain.
        
        :param types: a list of RDDL types
        :param msg: an error message to print in case the calculation fails.
        '''
        type_to_objects = self.type_to_objects
        try:
            return tuple(len(type_to_objects[ptype]) for ptype in types)
        except:
            for ptype in types:
                if ptype not in type_to_objects:
                    raise RDDLTypeError(
                        f'Type <{ptype}> is not valid, '
                        f'must be one of {set(type_to_objects.keys())}. {msg}')
    
    def ground_types(self, ptypes: Iterable[str]) -> Iterable[Tuple[str, ...]]:
        '''Given a list of valid types in the domain, produces an iterator
        of all possible assignments of objects to the types (groundings).
        
        Raises an exception if a type is invalid.
        '''
        if ptypes is None or not ptypes:
            return [()]
        objects_by_type = []
        for ptype in ptypes:
            objects = self.type_to_objects.get(ptype, None)
            if objects is None:
                raise RDDLTypeError(
                    f'Type <{ptype}> is not valid, '
                    f'must be one of {set(self.type_to_objects.keys())}.')
            objects_by_type.append(objects)
        return itertools.product(*objects_by_type)
            
    def ground_var_with_value(self, var: str, value: Value) -> Iterable[Tuple[str, Value]]:
        '''Converts a dictionary of var -> value associations to a dictionary
        of ground(var) -> value, where ground(var) is a grounding of var.
        '''
        # get the variable groundings
        groundings = self.variable_groundings.get(var, None)
        if groundings is None:
            raise RDDLTypeError(
                f'Variable <{var}> is not defined, '
                f'must be one of {set(self.variable_groundings.keys())}.')   
        
        return zip(groundings, itertools.repeat(value))
        
    def ground_vars_with_value(self, dict_values: Dict[str, Iterable[Value]]) -> Dict[str, Value]:
        '''Converts a dictionary of vars -> value associations to a dictionary 
        of ground(var) -> value, var is in vars, and ground(var) is a grounding 
        of var.
        '''
        grounded = {}
        for (var, value) in dict_values.items():
            grounded.update(self.ground_var_with_value(var, value))
        return grounded    
    
    def ground_var_with_values(self, var: str, values: Iterable[Value]) -> Iterable[Tuple[str, Value]]:
        '''Returns an iterator of (ground(var), value) pairs, where ground(var) is
        a grounding of var and value is the corresponding value in values.        
        '''
        # get the variable groundings
        groundings = self.variable_groundings.get(var, None)
        if groundings is None:
            raise RDDLTypeError(
                f'Variable <{var}> is not defined, '
                f'must be one of {set(self.variable_groundings.keys())}.')   
        
        # unravel the values and check size condition
        values = np.ravel(values, order='C')
        if len(groundings) != values.size:
            raise RDDLInvalidNumberOfArgumentsError(
                f'Variable <{var}> requires {len(groundings)} argument(s), '
                f'got {values.size}.')
            
        return zip(groundings, values)
    
    def ground_vars_with_values(self, dict_values: Dict[str, Iterable[Value]]) -> Dict[str, Value]:
        '''Converts a dictionary of vars -> values associations to a dictionary 
        of ground(var) -> value, var is in vars, ground(var) is a grounding of 
        var and value is the value in values.
        '''
        grounded = {}
        for (var, values) in dict_values.items():
            grounded.update(self.ground_var_with_values(var, values))
        return grounded
    
    def is_compatible(self, var: str, objects: List[str]) -> bool:
        '''Determines whether or not the given variable can be assigned the
        list of objects in the given order to its type parameters.
        '''
        ptypes = self.variable_params.get(var, None)
        if ptypes is None:
            return False
        if objects is None:
            objects = []
        if len(ptypes) != len(objects):
            return False
        for (ptype, obj) in zip(ptypes, objects):
            ptype_of_obj = self.object_to_type.get(obj, None)
            if ptype_of_obj is None or ptype != ptype_of_obj:
                return False
        return True
    
    def is_non_fluent_expression(self, expr: Expression) -> bool:
        '''Determines whether or not expression is a non-fluent expression
        (i.e. certified to produce the same value if evaluated multiple times
        with the same set of arguments). 
        
        Axiomatically, non-fluent expressions are:
            
            1. Non-expressions
            2. Constant expressions
            3. Free objects, i.e. ?x and enum objects, i.e. @x
            4. Non-fluent pvariables
            5. Expressions consisting of a deterministic operation on its
               child sub-expressions, if said children are all non-fluents.
        
        Note, that a random sample from a distribution is not non-fluent.
        '''
        if isinstance(expr, (tuple, list, set)):
            for arg in expr:
                if not self.is_non_fluent_expression(arg):
                    return False
            return True
        
        elif not isinstance(expr, Expression):
            return True
        
        etype, _ = expr.etype
        if etype == 'constant':
            return True
        
        elif etype == 'randomvar' or etype == 'randomvector':
            return False
        
        elif etype == 'pvar':
            name, pvars = expr.args
            
            # a free object (e.g., ?x) is non-fluent
            if RDDLPlanningModel.is_free_object(name):
                return True
                
            # an object is non-fluent
            elif not pvars and self.is_object(name):
                return True
                        
            # variable must be well-defined and non-fluent
            else:
                
                # check well-defined
                var_type = self.variable_types.get(name, None)     
                if var_type is None:
                    raise RDDLUndefinedVariableError(
                        f'Variable <{name}> is not defined, must be one of '
                        f'{set(self.variable_types.keys())}.')
                
                # check nested fluents
                if pvars is None:
                    pvars = []
                return var_type == 'non-fluent' \
                    and self.is_non_fluent_expression(pvars)
        
        else:
            return self.is_non_fluent_expression(expr.args)
    
    def expr_to_str(self):
        '''Returns a dictionary containing string representations of all 
        expressions in the current RDDL.
        '''
        printed = {
            'cpfs': {name: str(expr) for (name, (_, expr)) in self.cpfs.items()},
            'reward': str(self.reward),
            'invariants': [str(expr) for expr in self.invariants],
            'preconditions': [str(expr) for expr in self.preconditions],
            'terminations': [str(expr) for expr in self.terminations]
        }
        return printed
    
    def __str__(self):
        return pformat(vars(self))

    
class RDDLGroundedModel(RDDLPlanningModel):
    '''A class representing a RDDL domain + instance in grounded form.
    '''

    def __init__(self):
        super().__init__()
        
        self._variable_base_pvars = None
    
    @property
    def variable_base_pvars(self):
        return self._variable_base_pvars
    
    @variable_base_pvars.setter
    def variable_base_pvars(self, val):
        self._variable_base_pvars = val 



class RDDLLiftedModel(RDDLPlanningModel):
    '''A class representing a RDDL domain + instance in lifted form.
    '''
    
    def __init__(self, rddl):
        super(RDDLLiftedModel, self).__init__()
        
        self.ast = rddl
        
        self._extract_objects()            
        self._extract_variable_information()  
                
        self._extract_states()
        self._extract_non_fluents()
        self._extract_actions()
        self._extract_derived_and_interm()
        self._extract_observ()
        
        self.reward = self.ast.domain.reward
        self._extract_cpfs()
        self._extract_constraints()
        
        self._extract_horizon()
        self._extract_discount()
        self._extract_max_actions()
        
    def _extract_objects(self):
        
        # objects of each type as defined in the non-fluents {..} block
        ast_objects = self.ast.non_fluents.objects
        if not ast_objects or ast_objects[0] is None:
            ast_objects = []
        ast_objects = dict(ast_objects)
        
        # record the set of objects of each type defined in the domain
        objects, objects_rev, enums = {}, {}, set() 
        for (name, pvalues) in self.ast.domain.types:
            
            # check duplicated type
            if name in objects:
                raise RDDLInvalidObjectError(
                    f'Type <{name}> is repeated in types block.')
            
            # instance object
            if pvalues == 'object': 
                objects[name] = ast_objects.get(name, None)
                if objects[name] is None:
                    raise RDDLInvalidObjectError(
                        f'Type <{name}> has no object defined in the instance.')
                objects[name] = RDDLPlanningModel.strip_literals(objects[name])
                
            # domain object
            else: 
                objects[name] = RDDLPlanningModel.strip_literals(pvalues)
                enums.add(name)
        
            # make sure types do not share an object - record type of each object
            for obj in objects[name]:
                if obj in objects_rev:
                    if objects_rev[obj] == name:
                        raise RDDLInvalidObjectError(
                            f'Type <{name}> contains duplicated object <{obj}>.')
                    else:
                        raise RDDLInvalidObjectError(
                            f'Types <{name}> and <{objects_rev[obj]}> '
                            f'can not share the same object <{obj}>.')
                objects_rev[obj] = name
        
        # check that all types in instance are declared in domain
        for ptype in ast_objects:
            if ptype not in objects:
                raise RDDLInvalidObjectError(
                    f'Type <{ptype}> defined in the instance is not declared in '
                    f'the domain.')
        
        # maps each object to its canonical order as it appears in definition
        self.type_to_objects = objects
        self.object_to_type = objects_rev
        self.object_to_index = {obj: i 
                                for objs in objects.values() 
                                    for (i, obj) in enumerate(objs)}
        self.enum_types = enums
    
    def _extract_variable_information(self):
        var_types, var_ranges, var_params, var_default, var_ground = {}, {}, {}, {}, {}        
        for pvar in self.ast.domain.pvariables: 
            primed_name = name = pvar.name
            if pvar.is_state_fluent():
                primed_name = name + RDDLPlanningModel.NEXT_STATE_SYM
                
            # make sure variable is not defined more than once
            if name in var_params:
                raise RDDLRepeatedVariableError(
                    f'{pvar.fluent_type} <{name}> has the same name as '
                    f'another {var_types[name]} variable.')
                
            # make sure name does not contain separators
            for separator in (RDDLPlanningModel.FLUENT_SEP, RDDLPlanningModel.OBJECT_SEP):
                if separator in name:
                    raise RDDLInvalidObjectError(
                        f'Variable name <{name}> contains an '
                        f'illegal separator {separator}.')
            
            # record variable type
            var_types[name] = pvar.fluent_type
            if pvar.is_state_fluent():
                var_types[primed_name] = 'next-state-fluent'  
                              
            # record variable range
            var_ranges[name] = var_ranges[primed_name] = pvar.range 
            
            # record variable parameters
            ptypes = pvar.param_types
            if ptypes is None:
                ptypes = []
            var_params[name] = var_params[primed_name] = ptypes     
            
            # record default value
            var_default[name] = self._extract_default_value(pvar)
            
            # record possible groundings
            var_ground[name] = [
                RDDLPlanningModel.ground_var(name, objects)
                for objects in self.ground_types(ptypes)
            ]
            if pvar.is_state_fluent():
                var_ground[primed_name] = [
                    RDDLPlanningModel.ground_var(primed_name, objects)
                    for objects in self.ground_types(ptypes)
                ]
            
        self.variable_types = var_types
        self.variable_ranges = var_ranges
        self.variable_params = var_params
        self.variable_defaults = var_default
        self.variable_groundings = var_ground
        
    def _grounded_dict_to_dict_of_list(self, grounded_dict):
        new_dict = {}
        for (var, values_dict) in grounded_dict.items():
            grounded_names = list(values_dict.values())
            if self.variable_params[var]:
                new_dict[var] = grounded_names
            else:
                assert len(grounded_names) == 1
                new_dict[var] = grounded_names[0]
        return new_dict
    
    def _extract_default_value(self, pvar):
        prange, default = pvar.range, pvar.default
        if default is not None:
            
            # is an object
            if isinstance(default, str):
                default = RDDLPlanningModel.strip_literal(default)
                if default not in self.type_to_objects.get(prange, set()):
                    raise RDDLTypeError(
                        f'Default value {default} of variable <{pvar.name}> '
                        f'is not an object of type <{prange}>.')                     
            
            # is a primitive
            else:
                dtype = RDDLPlanningModel.PRIMITIVE_TYPES.get(prange, None)
                if dtype is None:
                    raise RDDLTypeError(
                        f'Type <{prange}> of variable <{pvar.name}> is an object '
                        f'or enumerated type, but assigned a default value '
                        f'{default}.')
                
                # type cast
                if not np.can_cast(default, dtype):
                    raise RDDLTypeError(
                        f'Default value {default} of variable <{pvar.name}> '
                        f'cannot be cast to required type <{prange}>.')
                default = dtype(default)
            
        return default
    
    def _extract_states(self):
        PRIME = RDDLPlanningModel.NEXT_STATE_SYM
        
        # get the information for each state from the domain
        states, statesranges, nextstates, prevstates = {}, {}, {}, {}
        for pvar in self.ast.domain.pvariables:
            if pvar.is_state_fluent():
                name = pvar.name
                statesranges[name] = pvar.range
                nextstates[name] = name + PRIME
                prevstates[name + PRIME] = name
                default = self.variable_defaults[name]
                states[name] = {gname: default
                                for gname in self.variable_groundings[name]}
                
        # update the state values with the values in the instance
        init_state_info = getattr(self.ast.instance, 'init_state', [])
        for ((name, params), value) in init_state_info:
                
            # check whether name is a valid state-fluent
            grounded_states = states.get(name, None)
            if grounded_states is None:
                raise RDDLUndefinedVariableError(
                    f'Variable <{name}> referenced in init-state block '
                    f'is not a valid state-fluent.')
                    
            # extract the grounded name and check that parameters are valid
            if params is not None:
                params = RDDLPlanningModel.strip_literals(params)
            gname = RDDLPlanningModel.ground_var(name, params)
            if gname not in grounded_states:
                required_types = self.variable_params[name]
                raise RDDLInvalidObjectError(
                    f'Parameter(s) {params} of state-fluent <{name}> '
                    f'declared in the init-state block are not valid, '
                    f'must be of type(s) {required_types}.')
                
            # make sure value is correct type
            if isinstance(value, str):
                value = RDDLPlanningModel.strip_literal(value)
                value_type = self.object_to_type.get(value, None)
                required_type = statesranges[name]
                if value_type != required_type:
                    if value_type is None:
                        raise RDDLInvalidObjectError(
                            f'State-fluent <{name}> of type <{required_type}> '
                            f'is initialized in init-state block with undefined '
                            f'object <{value}>.')
                    else:
                        raise RDDLInvalidObjectError(
                            f'State-fluent <{name}> of type <{required_type}> '
                            f'is initialized in init-state block with object '
                            f'<{value}> of type {value_type}.')
                        
            grounded_states[gname] = value
        
        self.state_fluents = self._grounded_dict_to_dict_of_list(states)  
        self.state_ranges = statesranges   
        self.next_state = nextstates
        self.prev_state = prevstates
    
    def _extract_non_fluents(self):
        
        # extract non-fluents values from the domain defaults
        non_fluents = {}
        for pvar in self.ast.domain.pvariables:
            if pvar.is_non_fluent():
                name = pvar.name
                default = self.variable_defaults[name] 
                non_fluents[name] = {gname: default
                                     for gname in self.variable_groundings[name]}
        
        # update non-fluent values with the values in the instance
        non_fluent_info = getattr(self.ast.non_fluents, 'init_non_fluent', [])
        for ((name, params), value) in non_fluent_info:
                
            # check whether name is a valid non-fluent
            grounded_names = non_fluents.get(name, None)
            if grounded_names is None:
                raise RDDLUndefinedVariableError(
                    f'Variable <{name}> referenced in non-fluents block '
                    f'is not a valid non-fluent.')
                
            # extract the grounded name and check that parameters are valid
            if params is not None:
                params = RDDLPlanningModel.strip_literals(params)
            gname = RDDLPlanningModel.ground_var(name, params)               
            if gname not in grounded_names:
                required_types = self.variable_params[name]
                raise RDDLInvalidObjectError(
                    f'Parameter(s) {params} of non-fluent <{name}> '
                    f'as declared in the non-fluents block are not valid, '
                    f'must be of type(s) {required_types}.')
                    
            # make sure value is correct type
            if isinstance(value, str):
                value = RDDLPlanningModel.strip_literal(value)
                value_type = self.object_to_type.get(value, None)
                required_type = self.variable_ranges[name]
                if value_type != required_type:
                    if value_type is None:
                        raise RDDLInvalidObjectError(
                            f'Non-fluent <{name}> of type <{required_type}> '
                            f'is initialized in non-fluents block with '
                            f'undefined object <{value}>.')
                    else:
                        raise RDDLInvalidObjectError(
                            f'Non-fluent <{name}> of type <{required_type}> '
                            f'is initialized in non-fluents block with object '
                            f'<{value}> of type <{value_type}>.')
                        
            grounded_names[gname] = value
                                        
        self.non_fluents = self._grounded_dict_to_dict_of_list(non_fluents)
    
    def _value_list_or_scalar_from_default(self, pvar):
        default = self.variable_defaults[pvar.name]
        ptypes = pvar.param_types   
        if ptypes is None:
            return default
        else:
            num_variations = 1
            for ptype in ptypes:
                num_variations *= len(self.type_to_objects[ptype])
            return [default] * num_variations
    
    def _extract_actions(self):
        actions, actionsranges = {}, {}
        for pvar in self.ast.domain.pvariables:
            if pvar.is_action_fluent():
                actionsranges[pvar.name] = pvar.range
                actions[pvar.name] = self._value_list_or_scalar_from_default(pvar)
        self.action_fluents = actions
        self.action_ranges = actionsranges
    
    def _extract_derived_and_interm(self):
        derived, interm = {}, {}
        for pvar in self.ast.domain.pvariables:
            if pvar.is_derived_fluent():
                derived[pvar.name] = self._value_list_or_scalar_from_default(pvar)
            elif pvar.is_intermediate_fluent():
                interm[pvar.name] = self._value_list_or_scalar_from_default(pvar)
        self.derived_fluents = derived
        self.interm_fluents = interm
    
    def _extract_observ(self):
        observ, observranges = {}, {}
        for pvar in self.ast.domain.pvariables:
            if pvar.is_observ_fluent():
                observranges[pvar.name] = pvar.range
                observ[pvar.name] = self._value_list_or_scalar_from_default(pvar)
        self.observ_fluents = observ
        self.observ_ranges = observranges
        
    def _extract_cpfs(self):
        cpfs = {}
        for cpf in self.ast.domain.cpfs[1]:
            name, objects = cpf.pvar[1]
            
            # make sure the CPF is defined in pvariables {...} scope
            types = self.variable_params.get(name, None)
            if types is None:
                raise RDDLUndefinedCPFError(
                    f'CPF <{name}> is not defined in pvariable block.')
            
            # make sure the CPF is not defined multiple times
            if name in cpfs:
                raise RDDLInvalidExpressionError(
                    f'Expression for CPF <{name}> is defined more than once '
                    'in the domain.')
                
            # make sure the number of parameters matches that in cpfs {...}
            if objects is None:
                objects = []
            if len(types) != len(objects):
                raise RDDLInvalidNumberOfArgumentsError(
                    f'Left-hand side of expression for CPF <{name}> requires '
                    f'{len(types)} parameter(s), got {objects}.')
            
            # check that the parameters are not literals
            for (index, pvar) in enumerate(objects):
                if not RDDLPlanningModel.is_free_object(pvar):
                    raise RDDLTypeError(
                        f'Definition for CPF <{name}> requires free '
                        f'object(s) on the left-hand side, but '
                        f'got the following expression at position {index + 1}:\n' + 
                        PST(pvar, f'CPF <{name}>'))
                
            # CPFs are stored as dictionary that associates cpf name with a pair 
            # the first element is the type argument list
            # the second element is the AST expression
            objects = list(zip(objects, types))
            cpfs[name] = (objects, cpf.expr)
        
        # make sure all CPFs have a valid expression in cpfs {...}
        for (var, fluent_type) in self.variable_types.items():
            if fluent_type in {'derived-fluent', 'interm-fluent',
                               'next-state-fluent', 'observ-fluent'} \
            and var not in cpfs:
                raise RDDLMissingCPFDefinitionError(
                    f'{fluent_type} CPF <{var}> is not defined in cpfs block.')
                    
        self.cpfs = cpfs
    
    def _extract_constraints(self):
        self.terminations = getattr(self.ast.domain, 'terminals', [])
        self.preconditions = getattr(self.ast.domain, 'preconds', [])
        self.invariants = getattr(self.ast.domain, 'invariants', [])

    def _extract_horizon(self):
        horizon = self.ast.instance.horizon
        if not (horizon >= 0):
            raise RDDLValueOutOfRangeError(
                f'Horizon {horizon} in the instance is not >= 0.')
        self.horizon = horizon

    def _extract_discount(self):
        discount = self.ast.instance.discount
        if not (0. <= discount):
            raise RDDLValueOutOfRangeError(
                f'Discount factor {discount} in the instance is not >= 0')
        self.discount = discount
        
    def _extract_max_actions(self):
        numactions = getattr(self.ast.instance, 'max_nondef_actions', 'pos-inf')
        if numactions == 'pos-inf':
            self.max_allowed_actions = sum(map(np.size, self.action_fluents.values()))
        else:
            self.max_allowed_actions = int(numactions)




from PIL import Image
import matplotlib
matplotlib.use('agg')
import matplotlib.pyplot as plt
import numpy as np

from pyRDDLGym.core.compiler.model import RDDLPlanningModel
from pyRDDLGym.core.visualizer.viz import BaseViz


class HeatmapVisualizer(BaseViz):

    def __init__(self, model: RDDLPlanningModel,
                 steps_history=None,
                 figure_size=(10, 10),
                 dpi=100,
                 fontsize=10,
                 cmap='seismic',
                 loccol='black',
                 scale_full_history: bool=True,
                 ranges=None) -> None:
        self._model = model
        self._figure_size = figure_size
        self._dpi = dpi
        self._fontsize = fontsize
        self._cmap = cmap
        self._loccol = loccol
        self._scale_full_history = scale_full_history
        self._ranges = ranges
        
        self._fig, self._ax = None, None
        self._data = None
        self._img = None
        
        if steps_history is None:
            steps_history = model.horizon
        self._steps = steps_history
        self._state_hist = {}
        self._state_shapes = {}
        self._labels = {}
        self._historic_min_max = {}
        for (state, values) in self._model.state_fluents.items():
            values = np.atleast_1d(values)
            self._state_hist[state] = np.full(shape=(len(values), self._steps),
                                              fill_value=np.nan)
            self._state_shapes[state] = self._model.object_counts(
                self._model.variable_params[state])
            self._labels[state] = list(map(
                ','.join, 
                self._model.ground_types(self._model.variable_params[state])
            ))
            self._historic_min_max[state] = (np.inf, -np.inf)
        self._step = 0
        self._drawn_artists = False
        
    def _setup(self):
        
        # prepare plot
        self._fig, self._ax = plt.subplots(len(self._state_hist), 1,
                                           squeeze=True,
                                           figsize=self._figure_size,
                                           sharex=True)
        if len(self._state_hist) == 1:
            self._ax = (self._ax,)
            
        # prepare artists
        self._lines = {}
        for (y, (state, values)) in enumerate(self._state_hist.items()):
            values = values[::-1, :]
            
            # title, labels and cursor line
            self._ax[y].xaxis.label.set_fontsize(self._fontsize)
            self._ax[y].yaxis.label.set_fontsize(self._fontsize)
            self._ax[y].title.set_text(state)
            self._ax[y].set_xlabel('decision epoch')
            self._ax[y].set_ylabel(state)
            
            im = self._ax[y].pcolormesh(values, 
                                        edgecolors=self._loccol, 
                                        linewidth=0.5, 
                                        cmap=self._cmap, 
                                        vmin=0.0, 
                                        vmax=0.0)         
            plt.colorbar(im, ax=self._ax[y])
            self._lines[state] = im
            
            labels = self._labels[state]
            self._ax[y].yaxis.set_ticks([0, len(labels)])
            self._ax[y].yaxis.set(ticks=np.arange(0.5, len(labels), 1), 
                                  ticklabels=labels) 
            self._ax[y].set_yticklabels(labels,
                                        fontdict={"fontsize": self._fontsize},
                                        rotation=30)
        plt.tight_layout()
                    
        self._backgrounds = [self._fig.canvas.copy_from_bbox(ax.bbox) for ax in self._ax]
        self._drawn_artists = True
        
    def convert2img(self, fig, ax):
        fig.canvas.draw()
        data = np.frombuffer(fig.canvas.tostring_rgb(), dtype=np.uint8)
        data = data.reshape(fig.canvas.get_width_height()[::-1] + (3,))
        img = Image.fromarray(data)
        self._data = data
        self._img = img
        return img

    def render(self, state):
        if not self._drawn_artists:
            self._setup()
            
        # update the state info
        if self._step >= self._steps:
            for values in self._state_hist.values():
                values[:,:-1] = values[:, 1:]
        states = {name: np.full(shape=shape, fill_value=np.nan)
                  for (name, shape) in self._state_shapes.items()}
        for (name, value) in state.items():
            if name not in self._model.variable_params:  # lifted
                var, objects = RDDLPlanningModel.parse_grounded(name)
                states[var][self._model.object_indices(objects)] = value
            else:  # grounded or scalar
                states[name] = value
        index = min(self._step, self._steps - 1)
        for (name, values) in states.items():
            self._state_hist[name][:, index] = np.ravel(values, order='C')
        
        # draw color plots
        for (y, (state, values)) in enumerate(self._state_hist.items()):
            values = values[::-1, :]
           
            # scaling of y axis
            vmin, vmax = None, None
            if self._scale_full_history:
                hmin, hmax = self._historic_min_max[state]
                valid_values = values[np.isfinite(values)]
                vmin = min(hmin, np.min(valid_values))
                vmax = max(hmax, np.max(valid_values))
                self._historic_min_max[state] = (vmin, vmax)            
            if self._ranges is not None and state in self._ranges:
                vmin, vmax = self._ranges[state]    
            
            # plot fluent as color mesh
            im = self._lines[state]
            self._fig.canvas.restore_region(self._backgrounds[y])
            self._lines[state].set_array(np.ravel(values))
            self._ax[y].draw_artist(self._lines[state])
            im.set_clim([vmin, vmax])
            
        self._step = self._step + 1
        
        img = self.convert2img(self._fig, self._ax)
        
        return img
    



'''In this example, a random policy is constructed and its performance is
evaluated on a specified domain. 

The syntax for running this example is:

    python run_gym.py <domain> <instance> [<episodes>] [<seed>]
    
where:
    <domain> is the name of a domain located in the /Examples directory
    <instance> is the instance number
    <episodes> is a positive integer for the number of episodes to simulate
    (defaults to 1)
    <seed> is a positive integer RNG key (defaults to 42)
'''
import sys

import pyRDDLGym
from pyRDDLGym.core.policy import RandomAgent


def main(domain, instance, episodes=1, seed=42):
    
    # create the environment
    env = pyRDDLGym.make(domain, instance, enforce_action_constraints=True)
    env.seed(seed)
    
    # set up a random policy
    agent = RandomAgent(action_space=env.action_space,
                        num_actions=env.max_allowed_actions,
                        seed=seed)
    agent.evaluate(env, episodes=episodes, verbose=True, render=True)
    
    # important when logging to save all traces
    env.close()


if __name__ == "__main__":
    args = sys.argv[1:]
    if len(args) < 2:
        print('python run_gym.py <domain> <instance> [<episodes>] [<seed>]')
        exit(1)
    kwargs = {'domain': args[0], 'instance': args[1]}
    if len(args) >= 3: kwargs['episodes'] = int(args[2])
    if len(args) >= 4: kwargs['seed'] = int(args[3])
    main(**kwargs)



import numpy as np
from typing import Dict, List, Set, Tuple, Union

from pyRDDLGym.core.compiler.model import RDDLPlanningModel
from pyRDDLGym.core.compiler.tracer import RDDLObjectsTracer, RDDLTracedObjects
from pyRDDLGym.core.debug.exception import (
    print_stack_trace as PST,
    RDDLInvalidNumberOfArgumentsError,
    RDDLInvalidObjectError,
    RDDLNotImplementedError,
    RDDLRepeatedVariableError,
    RDDLTypeError,
    RDDLUndefinedVariableError
)
from pyRDDLGym.core.debug.logger import Logger
from pyRDDLGym.core.parser.expr import Expression
from pyRDDLGym.core.simulator import lngamma


class RDDLIntervalAnalysis:
    
    def __init__(self, rddl: RDDLPlanningModel, trace: RDDLTracedObjects,
                 cpf_levels: Dict[str, List[str]],
                 logger: Logger=None):
        '''Creates a new interval analysis object for the given RDDL domain.
        
        :param rddl: the RDDL domain to analyze
        :param trace: pre-computed tracing of the RDDL
        :param cpf_levels: order of evaluation of CPFs
        :param logger: to log compilation information during tracing to file
        '''
        self.rddl = rddl
        self.trace = trace
        self.cpf_levels = cpf_levels
        self.logger = logger
        
        self.NUMPY_PROD_FUNC = np.frompyfunc(self._bound_product_scalar, nin=2, nout=1)    
        self.NUMPY_OR_FUNC = np.frompyfunc(self._bound_or_scalar, nin=2, nout=1)
    
    def bound(self) -> Dict[str, Tuple[np.ndarray, np.ndarray]]:
        rddl = self.rddl
        intervals = self._bound_initial_values()
        for _ in range(rddl.horizon):
            self._bound_next_epoch(intervals)
        return intervals
    
    def _bound_initial_values(self):
        rddl = self.rddl 
        
        # initially all bounds are calculated based on the initial values
        init_values = {}
        init_values.update(rddl.non_fluents)
        init_values.update(rddl.state_fluents)
        init_values.update(rddl.action_fluents)
        init_values.update(rddl.interm_fluents)
        init_values.update(rddl.derived_fluents)
        init_values.update(rddl.observ_fluents)
        
        # calculate initial values
        intervals = {}
        for (name, values) in init_values.items():
            params = rddl.variable_params[name]
            shape = rddl.object_counts(params)
            values = np.reshape(values, newshape=shape)
            intervals[name] = (values, values)
        return intervals
            
    def _bound_next_epoch(self, intervals):
        rddl = self.rddl 
        
        # trace and bound constraints
        for (i, expr) in enumerate(rddl.invariants):
            self._bound(expr, intervals)
        for (i, expr) in enumerate(rddl.preconditions):
            self._bound(expr, intervals)
            
        # trace and bound CPFs
        for cpfs in self.cpf_levels.values():
            for cpf in cpfs:
                _, expr = rddl.cpfs[cpf]
                lb0, ub0 = intervals.get(cpf, (None, None))
                lb1, ub1 = self._bound(expr, intervals)
                if cpf in rddl.interm_fluents or cpf in rddl.derived_fluents \
                or cpf in rddl.prev_state:
                    intervals[cpf] = (lb1, ub1)
                else:
                    intervals[cpf] = (np.minimum(lb0, lb1), np.maximum(ub0, ub1))
        
        # compute bounds on reward
        reward_bounds = self._bound(rddl.reward, intervals)
        intervals['reward'] = reward_bounds
        
        # update state bounds from next-state fluent bounds
        for (state, next_state) in rddl.next_state.items():
            intervals[state] = intervals[next_state]
                
    # ===========================================================================
    # start of bound calculation subroutines
    # ===========================================================================
    
    def _bound(self, expr, intervals):
        etype, _ = expr.etype
        if etype == 'constant':
            result = self._bound_constant(expr, intervals)
        elif etype == 'pvar':
            result = self._bound_pvar(expr, intervals)
        elif etype == 'arithmetic':
            result = self._bound_arithmetic(expr, intervals)
        elif etype == 'relational':
            result = self._bound_relational(expr, intervals)
        elif etype == 'boolean':
            result = self._bound_logical(expr, intervals)
        elif etype == 'aggregation':
            result = self._bound_aggregation(expr, intervals)
        elif etype == 'func':
            result = self._bound_func(expr, intervals)
        elif etype == 'control':
            result = self._bound_control(expr, intervals)
        elif etype == 'randomvar':
            result = self._bound_random(expr, intervals)
        elif etype == 'randomvector':
            result = self._bound_random_vector(expr, intervals)
        elif etype == 'matrix':
            result = self._bound_matrix(expr, intervals)
        else:
            raise RDDLNotImplementedError(
                f'Internal error: expression type {etype} is not supported.\n' + 
                PST(expr))
        
        # check valid bounds
        lower, upper = result
        if np.any(lower > upper):
            raise RuntimeError(
                f'Internal error: lower bound {lower} > upper bound {upper}.\n' +
                PST(expr))
        
        return result
    
    # ===========================================================================
    # leaves
    # ===========================================================================
        
    def _bound_constant(self, expr, intervals):
        lower = upper = self.trace.cached_sim_info(expr)
        return (lower, upper)
    
    def _bound_pvar(self, expr, intervals):
        var, args = expr.args
        
        # free variable (e.g., ?x) and object converted to canonical index
        is_value, cached_info = self.trace.cached_sim_info(expr)
        if is_value:
            lower = upper = cached_info
            return (lower, upper)
        
        # extract variable bounds
        bounds = intervals.get(var, None)
        if bounds is None:
            raise RDDLUndefinedVariableError(
                f'Variable <{var}> is referenced before assignment.\n' + PST(expr))
        lower, upper = bounds
        
        # propagate the bounds forward
        if cached_info is not None:
            slices, axis, shape, op_code, op_args = cached_info
            if slices: 
                raise RDDLNotImplementedError('Nested pvariable not supported.\n' + PST(expr))
            if axis:
                lower = np.expand_dims(lower, axis=axis)
                lower = np.broadcast_to(lower, shape=shape)
                upper = np.expand_dims(upper, axis=axis)
                upper = np.broadcast_to(upper, shape=shape)
            if op_code == RDDLObjectsTracer.NUMPY_OP_CODE.EINSUM:
                lower = np.einsum(lower, *op_args)
                upper = np.einsum(upper, *op_args)
            elif op_code == RDDLObjectsTracer.NUMPY_OP_CODE.TRANSPOSE:
                lower = np.transpose(lower, axes=op_args)
                upper = np.transpose(upper, axes=op_args)
        return (lower, upper)
    
    # ===========================================================================
    # arithmetic
    # ===========================================================================
    
    @staticmethod
    def _mask_assign(dest, mask, value, mask_value=False):
        assert (np.shape(dest) == np.shape(mask))
        if np.shape(dest):
            if mask_value:
                value = value[mask]
            dest[mask] = value
        elif mask:
            dest = value
        return dest
    
    @staticmethod
    def _bound_arithmetic_expr(int1, int2, op):
        (l1, u1), (l2, u2) = int1, int2
        
        # [a, b] + [c, d] = [a + c, b + d]
        if op == '+':
            lower = l1 + l2
            upper = u1 + u2
            return (lower, upper)
        
        # [a, b] - [c, d] = [a, d] - [b, c]
        elif op == '-':
            lower = l1 - u2
            upper = u1 - l2
            return (lower, upper)
        
        # [a, b] * [c, d] 
        elif op == '*':
            parts = [l1 * l2, l1 * u2, u1 * l2, u1 * u2]
            lower = np.minimum.reduce(parts)
            upper = np.maximum.reduce(parts)
            return (lower, upper)
        
        # [a, b] / [c, d] = [a, b] * (1 / [c, d])
        elif op == '/':
            mask_fn = RDDLIntervalAnalysis._mask_assign
            zero_in_open = (0 > l2) & (0 < u2)             
            l2_inv = 1. / u2
            u2_inv = 1. / l2
            l2_inv = mask_fn(l2_inv, u2 == 0, -np.inf)
            u2_inv = mask_fn(u2_inv, l2 == 0, +np.inf)
            l2_inv = mask_fn(l2_inv, zero_in_open, -np.inf)
            u2_inv = mask_fn(u2_inv, zero_in_open, +np.inf)
            int2_inv = (l2_inv, u2_inv)    
            return RDDLIntervalAnalysis._bound_arithmetic_expr(int1, int2_inv, '*')
        
        else:
            raise RDDLNotImplementedError(
                f'Arithmetic operator {op} is not supported.\n' + PST(expr))
        
    def _bound_arithmetic(self, expr, intervals):
        _, op = expr.etype
        args = expr.args        
        n = len(args)
        
        # unary negation
        if n == 1 and op == '-':
            arg, = args
            lower, upper = self._bound(arg, intervals)
            return (-upper, -lower)
        
        # binary operator
        elif n == 2:
            lhs, rhs = args
            int1 = self._bound(lhs, intervals)
            int2 = self._bound(rhs, intervals)
            return self._bound_arithmetic_expr(int1, int2, op)
        
        # ternary and higher order operator
        elif n >= 2 and op in ('+', '*'):
            int_res = self._bound(args[0], intervals)
            for arg in args[1:]:
                int_arg = self._bound(arg, intervals)
                int_res = self._bound_arithmetic_expr(int_res, int_arg, op)
            return int_res
        
        else:
            raise RDDLInvalidNumberOfArgumentsError(
                f'Arithmetic operator {op} does not have the required '
                f'number of arguments.\n' + PST(expr))
    
    # ===========================================================================
    # boolean
    # ===========================================================================
    
    @staticmethod
    def _bound_relational_expr(int1, int2, op):
        (l1, u1), (l2, u2) = int1, int2
        lower = np.zeros(np.shape(l1), dtype=int)
        upper = np.ones(np.shape(u1), dtype=int)
        mask_fn = RDDLIntervalAnalysis._mask_assign
        
        if op == '>=':
            lower = mask_fn(lower, l1 >= u2, 1)
            upper = mask_fn(upper, u1 < l2, 1)
            return lower, upper
        elif op == '>':
            lower = mask_fn(lower, l1 > u2, 1)
            upper = mask_fn(upper, u1 <= l2, 0)
            return lower, upper
        elif op == '==':
            lower = mask_fn(lower, (l1 == u1) & (l2 == u2) & (l1 == u2), 1)
            upper = mask_fn(upper, (u1 < l2) | (l1 > u2), 0)
            return lower, upper
        elif op == '~=':
            lower = mask_fn(lower, (u1 < l2) | (l1 > u2), 1)
            upper = mask_fn(upper, (l1 == u1) & (l2 == u2) & (l1 == u2), 0)
            return lower, upper
        elif op == '<':
            return RDDLIntervalAnalysis._bound_relational_expr(int2, int1, '>')
        elif op == '<=':
            return RDDLIntervalAnalysis._bound_relational_expr(int2, int1, '>=')
        else:
            raise RDDLNotImplementedError(
                f'Relational operator {op} is not supported.\n' + PST(expr))
         
    def _bound_relational(self, expr, intervals):
        _, op = expr.etype
        args = expr.args
        lhs, rhs = args
        int1 = self._bound(lhs, intervals)
        int2 = self._bound(rhs, intervals)
        return self._bound_relational_expr(int1, int2, op)
    
    @staticmethod
    def _bound_logical_expr(int1, int2, op):
        if op == '^':
            return RDDLIntervalAnalysis._bound_arithmetic_expr(int1, int2, '*')
        
        # x | y = ~(~x & ~y)
        elif op == '|':
            not1 = RDDLIntervalAnalysis._bound_logical_expr(int1, None, '~')
            not2 = RDDLIntervalAnalysis._bound_logical_expr(int2, None, '~')
            not12 = RDDLIntervalAnalysis._bound_logical_expr(not1, not2, '^')
            return RDDLIntervalAnalysis._bound_logical_expr(not12, None, '~')
        
        # x ~ y = (x | y) ^ ~(x & y)
        elif op == '~':
            if int2 is None:
                l1, u1 = int1
                lower = 1 - u1
                upper = 1 - l1
                return (lower, upper)
            else:
                or12 = RDDLIntervalAnalysis._bound_logical_expr(int1, int2, '|')
                and12 = RDDLIntervalAnalysis._bound_logical_expr(int1, int2, '^')
                not12 = RDDLIntervalAnalysis._bound_logical_expr(and12, None, '~')
                return RDDLIntervalAnalysis._bound_logical_expr(or12, not12, '^')
        
        # x => y = ~x | y
        elif op == '=>':
            not1 = RDDLIntervalAnalysis._bound_logical_expr(int1, None, '~')
            return RDDLIntervalAnalysis._bound_logical_expr(not1, int2, '|')
        
        # x <=> y = x == y
        elif op == '<=>':
            return RDDLIntervalAnalysis._bound_relational_expr(int1, int2, '==')
        
        else:
            raise RDDLNotImplementedError(
                f'Logical operator {op} is not supported.\n' + PST(expr))
           
    def _bound_logical(self, expr, intervals):
        _, op = expr.etype
        if op == '&':
            op = '^'
        args = expr.args
        n = len(args)
        
        # unary logical negation
        if n == 1:
            arg, = args
            int1 = self._bound(arg, intervals)
            return self._bound_logical_expr(int1, None, op)
        
        # binary operator
        elif n == 2:
            lhs, rhs = args
            int1 = self._bound(lhs, intervals)
            int2 = self._bound(rhs, intervals)
            return self._bound_logical_expr(int1, int2, op)
        
        # ternary and higher order operator
        elif n >= 2 and op in ('^', '|'):
            int_res = self._bound(args[0], intervals)
            for arg in args[1:]:
                int_arg = self._bound(arg, intervals)
                int_res = self._bound_logical_expr(int_res, int_arg, op)
            return int_res
        
        else:
            raise RDDLInvalidNumberOfArgumentsError(
                f'Logical operator {op} does not have the required '
                f'number of arguments.\n' + PST(expr))
        
    # ===========================================================================
    # aggregation
    # ===========================================================================
    
    @staticmethod
    def _bound_product_scalar(int1, int2):
        (l1, u1), (l2, u2) = int1, int2
        parts = [l1 * l2, l1 * u2, u1 * l2, u1 * u2]
        lower = min(parts)
        upper = max(parts)
        return (lower, upper)
        
    @staticmethod
    def _bound_or_scalar(int1, int2):
        (l1, u1), (l2, u2) = int1, int2
        l1, u1 = (1 - u1, 1 - l1)
        l2, u2 = (1 - u2, 1 - l2)        
        parts = [l1 * l2, l1 * u2, u1 * l2, u1 * u2]
        not_l, not_u = min(parts), max(parts)        
        lower = 1 - not_u
        upper = 1 - not_l
        return (lower, upper)        
    
    @staticmethod
    def _zip_bounds_to_single_array(lower, upper):
        shape = np.shape(lower)
        lower = np.ravel(lower, order='C')
        upper = np.ravel(upper, order='C')
        bounds = np.empty(lower.size, dtype=object)
        bounds[:] = list(zip(lower, upper))
        bounds = np.reshape(bounds, newshape=shape, order='C')
        return bounds
    
    @staticmethod
    def _unzip_single_array_to_bounds(array):
        if isinstance(array, tuple):
            return array
        shape = np.shape(array)
        array = np.ravel(array, order='C')
        lower = np.reshape([p[0] for p in array], newshape=shape, order='C')
        upper = np.reshape([p[1] for p in array], newshape=shape, order='C')
        return (lower, upper)
    
    @staticmethod
    def _bound_aggregation_func(lower, upper, axes, numpyfunc):
        array = RDDLIntervalAnalysis._zip_bounds_to_single_array(lower, upper)
        for axis in sorted(axes, reverse=True):
            array = numpyfunc.reduce(array, axis=axis)
        return RDDLIntervalAnalysis._unzip_single_array_to_bounds(array)
        
    def _bound_aggregation(self, expr, intervals):
        _, op = expr.etype
        * _, arg = expr.args        
        lower, upper = self._bound(arg, intervals)
        
        # aggregate bound calculation over axes
        _, axes = self.trace.cached_sim_info(expr)
        if op == 'sum':
            return (np.sum(lower, axis=axes), np.sum(upper, axis=axes))        
        elif op == 'avg':
            return (np.mean(lower, axis=axes), np.mean(upper, axis=axes))        
        elif op == 'prod':
            return self._bound_aggregation_func(lower, upper, axes, self.NUMPY_PROD_FUNC)        
        elif op == 'minimum':
            return (np.min(lower, axis=axes), np.min(upper, axis=axes))        
        elif op == 'maximum':
            return (np.max(lower, axis=axes), np.max(upper, axis=axes))        
        elif op == 'forall':
            return self._bound_aggregation_func(lower, upper, axes, self.NUMPY_PROD_FUNC)        
        elif op == 'exists':
            return self._bound_aggregation_func(lower, upper, axes, self.NUMPY_OR_FUNC)
        else:        
            raise RDDLNotImplementedError(
                f'Aggregation operator {op} is not supported.\n' + PST(expr))
    
    # ===========================================================================
    # function
    # ===========================================================================
    
    UNARY_MONOTONE = {
        'sgn': lambda x: np.sign(x).astype(int),
        'round': lambda x: np.round(x).astype(int),
        'floor': lambda x: np.floor(x).astype(int),
        'ceil': lambda x: np.ceil(x).astype(int),
        'acos': np.arccos,
        'asin': np.arcsin,
        'atan': np.arctan,
        'sinh': np.sinh,
        'tanh': np.tanh,
        'exp': np.exp,
        'ln': np.log,
        'sqrt': np.sqrt
    }
    
    UNARY_U_SHAPED = {
        'abs': (np.abs, 0),
        'cosh': (np.cosh, 0),
        'lngamma': (lngamma, 1.4616321449683623),
        'gamma': ((lambda x: np.exp(lngamma(x))), 1.4616321449683623)
    }
    
    @staticmethod
    def _bound_func_monotone(l, u, func):
        fl, fu = func(l), func(u)
        lower = np.minimum(fl, fu)
        upper = np.maximum(fl, fu)
        return (lower, upper)
    
    @staticmethod
    def _next_multiple(start, offset, period=2 * np.pi):
        multiples = -np.floor_divide(start - offset, -period)
        next_up = period * multiples + offset
        return next_up
        
    @staticmethod
    def _bound_func_u_shaped(l, u, x_crit, func):
        fl, fu = func(l), func(u)
        f_crit = func(x_crit)
        lower = np.full(shape=np.shape(l), fill_value=f_crit)
        upper = np.maximum(fl, fu)
        lower = RDDLIntervalAnalysis._mask_assign(lower, l >= x_crit, fl, True)
        upper = RDDLIntervalAnalysis._mask_assign(upper, l >= x_crit, fu, True)
        lower = RDDLIntervalAnalysis._mask_assign(lower, u <= x_crit, fu, True)
        upper = RDDLIntervalAnalysis._mask_assign(upper, u <= x_crit, fl, True)
        return (lower, upper)
    
    def _bound_func_unary(self, expr, intervals):
        _, name = expr.etype
        arg, = expr.args
        l, u = self._bound(arg, intervals)
        
        # f is monotone, x in [a, b] => f(x) in [f(a), f(b)]
        if name in self.UNARY_MONOTONE:
            return self._bound_func_monotone(l, u, self.UNARY_MONOTONE[name])
        
        # f is a u-shaped function
        elif name in self.UNARY_U_SHAPED:
            func, x_crit = self.UNARY_U_SHAPED[name]
            return self._bound_func_u_shaped(l, u, x_crit, func)
        
        # special functions sin
        elif name == 'sin':
            
            # by default evaluate at end points
            sin_l, sin_u = np.sin(l), np.sin(u)
            lower = np.minimum(sin_l, sin_u)
            upper = np.maximum(sin_l, sin_u)
            
            # any infinity results in [-1, 1]
            has_inf = (l == -np.inf) | (u == np.inf)
            lower = self._mask_assign(lower, has_inf, -1.0)
            upper = self._mask_assign(upper, has_inf, +1.0)
            
            # find critical points
            next_max = self._next_multiple(l, np.pi / 2)
            next_min = self._next_multiple(l, 3 * np.pi / 2)
            lower = self._mask_assign(lower, next_min <= u, -1.0)
            upper = self._mask_assign(upper, next_max <= u, +1.0)
            return lower, upper
        
        # special functions cos
        elif name == 'cos':
            
            # by default evaluate at end points
            cos_l, cos_u = np.cos(l), np.cos(u)
            lower = np.minimum(cos_l, cos_u)
            upper = np.maximum(cos_l, cos_u)
            
            # any infinity results in [-1, 1]
            has_inf = (l == -np.inf) | (u == np.inf)
            lower = self._mask_assign(lower, has_inf, -1.0)
            upper = self._mask_assign(upper, has_inf, +1.0)
            
            # find critical points
            next_max = self._next_multiple(l, 0)
            next_min = self._next_multiple(l, np.pi)
            lower = self._mask_assign(lower, next_min <= u, -1.0)
            upper = self._mask_assign(upper, next_max <= u, +1.0)
            return lower, upper
        
        # special functions tan
        elif name == 'tan':
            
            # by default evaluate at end points
            tan_l, tan_u = np.tan(l), np.tan(u)
            lower = np.minimum(tan_l, tan_u)
            upper = np.maximum(tan_l, tan_u)
            
            # any infinity results in [-inf, inf]
            has_inf = (l == -np.inf) | (u == np.inf)
            lower = self._mask_assign(lower, has_inf, -np.inf)
            upper = self._mask_assign(upper, has_inf, +np.inf)
            
            # an asymptote is inside the open interval, set to [-inf, inf]
            next_asymptote = self._next_multiple(l, np.pi / 2, np.pi)
            asymptote_in_open = (l < next_asymptote) & (u > next_asymptote)
            lower = self._mask_assign(lower, asymptote_in_open, -np.inf)
            upper = self._mask_assign(upper, asymptote_in_open, +np.inf)
            
            # an asymptote is on an edge
            lower = self._mask_assign(lower, l == next_asymptote, -np.inf)
            upper = self._mask_assign(upper, u == next_asymptote, +np.inf)
            upper = self._mask_assign(upper, u == next_asymptote + np.pi, +np.inf)            
            return lower, upper
        
        else:
            raise RDDLNotImplementedError(
                f'Unary function {name} is not supported.\n' + PST(expr))
    
    @staticmethod
    def _bound_func_power(int1, int2):
        (l1, u1), (l2, u2) = int1, int2        
        lower = np.full(shape=np.shape(l1), fill_value=-np.inf)
        upper = np.full(shape=np.shape(u1), fill_value=+np.inf)
        mask_fn = RDDLIntervalAnalysis._mask_assign
        
        # positive base, means well defined for any real power
        log1 = RDDLIntervalAnalysis._bound_func_monotone(l1, u1, np.log)
        pow = RDDLIntervalAnalysis._bound_arithmetic_expr(log1, int2, '*')
        l1pos, u1pos = RDDLIntervalAnalysis._bound_func_monotone(*pow, np.exp)
        lower = mask_fn(lower, l1 > 0, l1pos, True)
        upper = mask_fn(upper, l1 > 0, u1pos, True)
            
        # otherwise, defined if the power is an integer point
        l2_is_int = np.equal(np.mod(l2, 1), 0)
        u2_is_int = np.equal(np.mod(u2, 1), 0)
        pow_valid = l2_is_int & u2_is_int & (l2 >= 0) & (l2 == u2)
        pow = l2.astype(int) if np.shape(l2) else int(l2)
        pow_even = (np.mod(pow, 2) == 0)
            
        # if the base interval contains 0
        case1 = pow_valid & (0 >= l1) & (0 <= u1)
        lower = mask_fn(lower, case1 & pow_even, 0)
        upper = mask_fn(upper, case1 & (pow == 0), 1)
        upper = mask_fn(upper, case1 & (pow > 0) & pow_even, 
                        np.maximum(l1 ** pow, l2 ** pow), True)
        lower = mask_fn(lower, case1 & ~pow_even, l1 ** pow, True)
        upper = mask_fn(upper, case1 & ~pow_even, l2 ** pow, True)
            
        # if the base is strictly negative
        case2 = pow_valid & (u1 < 0)
        lower = mask_fn(lower, case2 & (pow == 0), 1)
        upper = mask_fn(upper, case2 & (pow == 0), 1)
        lower = mask_fn(lower, case2 & pow_even, u1 ** pow, True)
        upper = mask_fn(upper, case2 & pow_even, l1 ** pow, True)
        lower = mask_fn(lower, case2 & ~pow_even, l1 ** pow, True)
        upper = mask_fn(upper, case2 & ~pow_even, u1 ** pow, True)
        return lower, upper
    
    def _bound_func_binary(self, expr, intervals):
        _, name = expr.etype
        arg1, arg2 = expr.args
        int1 = (l1, u1) = self._bound(arg1, intervals)
        int2 = (l2, u2) = self._bound(arg2, intervals)
        
        # div(x, y) = floor(x / y)
        if name == 'div': 
            l, u = self._bound_arithmetic_expr(int1, int2, '/')
            return self._bound_func_monotone(l, u, np.floor)
        
        # min([a, b], [c, d]) = (min[a, c], min[b, d])
        elif name == 'min':
            return (np.minimum(l1, l2), np.minimum(u1, u2))
        
        # max([a, b], [c, d]) = (max[a, c], max[b, d])
        elif name == 'max':
            return (np.maximum(l1, l2), np.maximum(u1, u2))
        
        # TODO
        elif name == 'mod':
            raise RDDLNotImplementedError(
                f'Binary function {name} is not supported.\n' + PST(expr))
        
        # TODO
        elif name == 'fmod':
            raise RDDLNotImplementedError(
                f'Binary function {name} is not supported.\n' + PST(expr))
        
        elif name == 'pow':
            return self._bound_func_power(int1, int2)
        
        # log[x, b] = log[x] / log[b]
        elif name == 'log':
            if np.any(l2 <= 0):
                raise RDDLNotImplementedError(
                    f'Function {name} with base <= 0 is not supported.\n' + PST(expr))
            if np.any(l1 <= 0):
                raise RDDLNotImplementedError(
                    f'Function {name} with argument <= 0 is not supported.\n' + PST(expr))                    
            log1 = self._bound_func_monotone(l1, u1, np.log)
            log2 = self._bound_func_monotone(l2, u2, np.log)
            return self._bound_arithmetic_expr(log1, log2, '/')
        
        # hypot[x, y] = sqrt[x * x + y * y]
        elif name == 'hypot': 
            pow1 = self._bound_func_u_shaped(l1, u1, 0, np.square)
            pow2 = self._bound_func_u_shaped(l2, u2, 0, np.square)
            l, u = self._bound_arithmetic_expr(pow1, pow2, '+')
            return self._bound_func_monotone(l, u, np.sqrt)
        
        else:
            raise RDDLNotImplementedError(
                f'Binary function {name} is not supported.\n' + PST(expr))
        
    def _bound_func(self, expr, intervals):
        _, name = expr.etype
        args = expr.args
        n = len(args)
        
        if n == 1:
            return self._bound_func_unary(expr, intervals)
        elif n == 2:
            return self._bound_func_binary(expr, intervals)
        else:
            raise RDDLNotImplementedError(
                f'Function {name} with {n} arguments is not supported.\n' + 
                PST(expr))
            
    # ===========================================================================
    # control flow
    # ===========================================================================
    
    def _bound_control(self, expr, intervals):
        _, op = expr.etype
        
        if op == 'if':
            return self._bound_if(expr, intervals)
        else:
            return self._bound_switch(expr, intervals) 
    
    def _bound_if(self, expr, intervals):
        args = expr.args
        pred, arg1, arg2 = args
        intp = (lp, up) = self._bound(pred, intervals)
        int1 = (l1, u1) = self._bound(arg1, intervals)
        int2 = (l2, u2) = self._bound(arg2, intervals)      
        lower = np.minimum(l1, l2)
        upper = np.maximum(u1, u2)
        
        # reduce range if the predicate is known with certainty
        lower = self._mask_assign(lower, lp >= 1, l1, True)
        upper = self._mask_assign(upper, lp >= 1, u1, True)
        lower = self._mask_assign(lower, up <= 0, l2, True)
        upper = self._mask_assign(upper, up <= 0, u2, True)
        return (lower, upper)
            
    def _bound_switch(self, expr, intervals):
        raise RDDLNotImplementedError(
            f'Function switch is not supported.\n' + PST(expr))
    
    # ===========================================================================
    # random variables
    # ===========================================================================
    
    def _bound_random(self, expr, intervals):
        _, name = expr.etype
        if name == 'KronDelta':
            return self._bound_random_kron(expr, intervals)
        elif name == 'DiracDelta':
            return self._bound_random_dirac(expr, intervals)
        elif name == 'Uniform':
            return self._bound_uniform(expr, intervals)            
        elif name == 'Bernoulli':
            return self._bound_bernoulli(expr, intervals)
        elif name == 'Normal':
            return self._bound_normal(expr, intervals)
        elif name == 'Poisson':
            return self._bound_poisson(expr, intervals)
        elif name == 'Exponential':
            return self._bound_exponential(expr, intervals)
        elif name == 'Weibull':
            return self._bound_weibull(expr, intervals)        
        elif name == 'Gamma':
            return self._bound_gamma(expr, intervals)
        elif name == 'Binomial':
            return self._bound_binomial(expr, intervals)
        elif name == 'NegativeBinomial':
            return self._bound_negative_binomial(expr, intervals)
        elif name == 'Beta':
            return self._bound_beta(expr, intervals)
        elif name == 'Geometric':
            return self._bound_geometric(expr, intervals)
        elif name == 'Pareto':
            return self._bound_pareto(expr, intervals)
        elif name == 'Student':
            return self._bound_student(expr, intervals)
        elif name == 'Gumbel':
            return self._bound_gumbel(expr, intervals)
        elif name == 'Laplace':
            return self._bound_laplace(expr, intervals)
        elif name == 'Cauchy':
            return self._bound_cauchy(expr, intervals)
        elif name == 'Gompertz':
            return self._bound_gompertz(expr, intervals)
        elif name == 'ChiSquare':
            return self._bound_chisquare(expr, intervals)
        elif name == 'Kumaraswamy':
            return self._bound_kumaraswamy(expr, intervals)
        elif name == 'Discrete':
            return self._bound_discrete(expr, intervals, False)
        elif name == 'UnnormDiscrete':
            return self._bound_discrete(expr, intervals, True)
        elif name == 'Discrete(p)':
            return self._bound_discrete_pvar(expr, intervals, False)
        elif name == 'UnnormDiscrete(p)':
            return self._bound_discrete_pvar(expr, intervals, True)
        else:
            raise RDDLNotImplementedError(
                f'Distribution {name} is not supported.\n' + PST(expr))
    
    def _bound_random_kron(self, expr, intervals):
        args = expr.args
        arg, = args
        return self._bound(arg, intervals)
    
    def _bound_random_dirac(self, expr, intervals):
        args = expr.args
        arg, = args
        return self._bound(arg, intervals)
    
    def _bound_uniform(self, expr, intervals):
        args = expr.args
        a, b = args
        intl = (la, ua) = self._bound(a, intervals)
        intu = (lb, ub) = self._bound(b, intervals)
        lower = np.minimum(la, ua)
        upper = np.maximum(lb, ub)
        return (lower, upper)
    
    def _bound_bernoulli(self, expr, intervals):
        args = expr.args
        p, = args
        intp = (lp, up) = self._bound(p, intervals)
        
        lower = np.zeros(shape=np.shape(lp), dtype=int)
        upper = np.ones(shape=np.shape(up), dtype=int)
        lower = self._mask_assign(lower, lp >= 1, 1)
        upper = self._mask_assign(upper, up <= 0, 0)
        return (lower, upper)
    
    def _bound_normal(self, expr, intervals):
        args = expr.args
        mean, var = args
        intm = (lm, um) = self._bound(mean, intervals)
        intv = (lv, uv) = self._bound(var, intervals)
        
        lower = np.full(shape=np.shape(lm), fill_value=-np.inf)
        upper = np.full(shape=np.shape(um), fill_value=+np.inf)
        lower = self._mask_assign(lower, (lv == 0) & (uv == 0), lm, True)
        upper = self._mask_assign(upper, (lv == 0) & (uv == 0), um, True)
        return (lower, upper)
    
    def _bound_poisson(self, expr, intervals):
        args = expr.args
        p, = args
        intp = (lp, up) = self._bound(p, intervals)
        
        lower = np.zeros(shape=np.shape(lp), dtype=int)
        upper = np.full(shape=np.shape(up), fill_value=np.inf)
        return (lower, upper)
    
    def _bound_exponential(self, expr, intervals):
        args = expr.args
        scale, = args
        ints = (ls, us) = self._bound(scale, intervals)
        
        lower = np.zeros(shape=np.shape(ls))
        upper = np.full(shape=np.shape(us), fill_value=np.inf)
        return (lower, upper)
    
    def _bound_weibull(self, expr, intervals):
        args = expr.args
        shape, scale = args
        intsh = (lsh, ush) = self._bound(shape, intervals)
        intsc = (lsc, usc) = self._bound(scale, intervals)
        
        lower = np.zeros(shape=np.shape(lsh))
        upper = np.full(shape=np.shape(ush), fill_value=np.inf)
        return (lower, upper)
    
    def _bound_gamma(self, expr, intervals):
        args = expr.args
        shape, scale = args
        intsh = (lsh, ush) = self._bound(shape, intervals)
        intsc = (lsc, usc) = self._bound(scale, intervals)
        
        lower = np.zeros(shape=np.shape(ls))
        upper = np.full(shape=np.shape(us), fill_value=np.inf)
        return (lower, upper)
    
    def _bound_binomial(self, expr, intervals):
        args = expr.args
        n, p = args
        intn = (ln, un) = self._bound(n, intervals)
        intp = (lp, up) = self._bound(p, intervals)
        
        lower = np.zeros(shape=np.shape(ln), dtype=int)
        upper = un
        return (lower, upper)
    
    def _bound_beta(self, expr, intervals):
        args = expr.args
        shape, rate = args
        ints = (ls, us) = self._bound(shape, intervals)
        intr = (lr, ur) = self._bound(rate, intervals)
        
        lower = np.zeros(shape=np.shape(ls))
        upper = np.ones(shape=np.shape(us))
        return (lower, upper)
    
    def _bound_geometric(self, expr, intervals):
        args = expr.args
        p, = args
        
        intp = (lp, up) = self._bound(p, intervals)
        lower = np.ones(shape=np.shape(lp), dtype=int)
        upper = np.full(shape=np.shape(up), fill_value=np.inf)
        return (lower, upper)
    
    def _bound_pareto(self, expr, intervals):
        args = expr.args
        shape, scale = args
        intsh = (lsh, ush) = self._bound(shape, intervals)
        intsc = (lsc, usc) = self._bound(scale, intervals)
        
        lower = lsc
        upper = np.full(shape=np.shape(usc), fill_value=np.inf)
        return (lower, upper)
    
    def _bound_student(self, expr, intervals):
        args = expr.args
        df, = args
        intd = (ld, ud) = self._bound(df, intervals)
        
        lower = np.full(shape=np.shape(ld), fill_value=-np.inf)
        upper = np.full(shape=np.shape(ud), fill_value=+np.inf)
        return (lower, upper)
    
    def _bound_gumbel(self, expr, intervals):
        args = expr.args
        mean, scale = args
        intm = (lm, um) = self._bound(mean, intervals)
        ints = (ls, us) = self._bound(scale, intervals)
        
        lower = np.full(shape=np.shape(lm), fill_value=-np.inf)
        upper = np.full(shape=np.shape(um), fill_value=+np.inf)
        return (lower, upper)
    
    def _bound_cauchy(self, expr, intervals):
        args = expr.args
        mean, scale = args
        intm = (lm, um) = self._bound(mean, intervals)
        ints = (ls, us) = self._bound(scale, intervals)
        
        lower = np.full(shape=np.shape(lm), fill_value=-np.inf)
        upper = np.full(shape=np.shape(um), fill_value=+np.inf)
        return (lower, upper)
    
    def _bound_gompertz(self, expr, intervals):
        args = expr.args
        shape, scale = args
        intsh = (lsh, ush) = self._bound(shape, intervals)
        intsc = (lsc, usc) = self._bound(scale, intervals)
        
        lower = np.zeros(shape=np.shape(lsh))
        upper = np.full(shape=np.shape(ush), fill_value=np.inf)
        return (lower, upper)
    
    def _bound_chisquare(self, expr, intervals):
        args = expr.args
        df, = args
        intd = (ld, ud) = self._bound(df, intervals)
        
        lower = np.zeros(shape=np.shape(ld))
        upper = np.full(shape=np.shape(ud), fill_value=np.inf)
        return (lower, upper)
    
    def _bound_kumaraswamy(self, expr, intervals):
        args = expr.args
        a, b = args
        inta = (la, ua) = self._bound(a, intervals)
        intb = (lb, ub) = self._bound(b, intervals)
        
        lower = np.zeros(shape=np.shape(la))
        upper = np.ones(shape=np.shape(ua))
        return (lower, upper)



# This file is based on thiago pbueno's pyrddl.
# https://github.com/thiagopbueno/pyrddl
# it was adapted and extended for pyRDDLGym

from typing import Dict, List, Sequence, Tuple, Union

FluentTuple = Tuple[Union[str, None]]
Value = Union[bool, int, float]
FluentInitializer = Tuple[FluentTuple, Value]
FluentInitializerList = List[FluentInitializer]

ObjectsList = List[Tuple[str, List[str]]]


class NonFluents(object):
    '''NonFluents class for accessing RDDL non-fluents sections.
    Note:
        This class is intended to be solely used by the parser and compiler.
        Do not attempt to directly use this class to build a NonFluents object.
    Args:
        name: Name of RDDL non-fluents.
        sections: Mapping from string to non-fluents section.
    Attributes:
        name (str): Name of RDDL non-fluents block.
        domain (str): Name of RDDL domain block.
        objects (:obj:`ObjectsList`): List of RDDL objects for each type.
        init_non_fluent (:obj:`FluentInitializerList`): List of non-fluent initializers.
    '''

    def __init__(self, name: str, sections: Dict[str, Sequence]) -> None:
        self.name = name
        self.objects = []
        self.__dict__.update(sections)



# This file is part of thiago pbueno's pyrddl.
# https://github.com/thiagopbueno/pyrddl

import collections
import itertools
from typing import Dict, List, Sequence, Optional, Tuple, Union

from pyRDDLGym.core.parser.domain import Domain
from pyRDDLGym.core.parser.instance import Instance
from pyRDDLGym.core.parser.nonfluents import NonFluents

Block = Union[Domain, NonFluents, Instance]
ObjectStruct = Dict[str, Union[int, Dict[str, int], List[str]]]
ObjectTable = Dict[str, ObjectStruct]
FluentParamsList = Sequence[Tuple[str, List[str]]]


class RDDL(object):
    '''RDDL class for accessing RDDL blocks.
    Note:
        This class is intended to be solely used by the parser and compiler.
        Do not attempt to directly use this class to build a RDDL object.
    Args:
        blocks: Mapping from string to RDDL block.
    Attributes:
        domain (:obj:`Domain`): RDDL domain block.
        non_fluents (:obj:`NonFluents`): RDDL non-fluents block.
        instance (:obj:`Instance`): RDDL instance block.
        object_table (:obj:`ObjectTable`): The object table for each RDDL type.
    '''

    def __init__(self, blocks: Dict[str, Block]) -> None:
        self.domain = blocks['domain']
        self.non_fluents = blocks['non_fluents']
        self.instance = blocks['instance']

    def build(self):
        self.domain.build()
        self._build_object_table()
        self._build_fluent_table()

    def _build_object_table(self):
        '''Builds the object table for each RDDL type.'''
        types = self.domain.types
        objects = dict(self.non_fluents.objects)
        self.object_table = dict()
        for name, value in self.domain.types:
            if value == 'object':
                objs = objects[name]
                idx = { obj: i for i, obj in enumerate(objs) }
                self.object_table[name] = {
                    'size': len(objs),
                    'idx': idx,
                    'objects': objs
                }

    def _build_fluent_table(self):
        '''Builds the fluent table for each RDDL pvariable.'''
        self.fluent_table = collections.OrderedDict()

        for name, size in zip(self.domain.non_fluent_ordering, self.non_fluent_size):
            non_fluent = self.domain.non_fluents[name]
            self.fluent_table[name] = (non_fluent, size)

        for name, size in zip(self.domain.state_fluent_ordering, self.state_size):
            fluent = self.domain.state_fluents[name]
            self.fluent_table[name] = (fluent, size)

        for name, size in zip(self.domain.action_fluent_ordering, self.action_size):
            fluent = self.domain.action_fluents[name]
            self.fluent_table[name] = (fluent, size)

        for name, size in zip(self.domain.interm_fluent_ordering, self.interm_size):
            fluent = self.domain.intermediate_fluents[name]
            self.fluent_table[name] = (fluent, size)

    @property
    def non_fluent_variables(self) -> FluentParamsList:
        '''Returns the instantiated non-fluents in canonical order.
        Returns:
            Sequence[Tuple[str, List[str]]]: A tuple of pairs of fluent name
            and a list of instantiated fluents represented as strings.
        '''
        fluents = self.domain.non_fluents
        ordering = self.domain.non_fluent_ordering
        return self._fluent_params(fluents, ordering)

    @property
    def state_fluent_variables(self) -> FluentParamsList:
        '''Returns the instantiated state fluents in canonical order.
        Returns:
            Sequence[Tuple[str, List[str]]]: A tuple of pairs of fluent name
            and a list of instantiated fluents represented as strings.
        '''
        fluents = self.domain.state_fluents
        ordering = self.domain.state_fluent_ordering
        return self._fluent_params(fluents, ordering)

    @property
    def interm_fluent_variables(self) -> FluentParamsList:
        '''Returns the instantiated intermediate fluents in canonical order.
        Returns:
            Sequence[Tuple[str, List[str]]]: A tuple of pairs of fluent name
            and a list of instantiated fluents represented as strings.
        '''
        fluents = self.domain.intermediate_fluents
        ordering = self.domain.interm_fluent_ordering
        return self._fluent_params(fluents, ordering)

    @property
    def action_fluent_variables(self) -> FluentParamsList:
        '''Returns the instantiated action fluents in canonical order.
        Returns:
            Sequence[Tuple[str, List[str]]]: A tuple of pairs of fluent name
            and a list of instantiated fluents represented as strings.
        '''
        fluents = self.domain.action_fluents
        ordering = self.domain.action_fluent_ordering
        return self._fluent_params(fluents, ordering)

    @property
    def non_fluent_size(self) -> Sequence[Sequence[int]]:
        '''The size of each non-fluent in canonical order.
        Returns:
            Sequence[Sequence[int]]: A tuple of tuple of integers
            representing the shape and size of each non-fluent.
        '''
        fluents = self.domain.non_fluents
        ordering = self.domain.non_fluent_ordering
        return self._fluent_size(fluents, ordering)

    @property
    def state_size(self) -> Sequence[Sequence[int]]:
        '''The size of each state fluent in canonical order.
        Returns:
            Sequence[Sequence[int]]: A tuple of tuple of integers
            representing the shape and size of each fluent.
        '''
        fluents = self.domain.state_fluents
        ordering = self.domain.state_fluent_ordering
        return self._fluent_size(fluents, ordering)

    @property
    def action_size(self) -> Sequence[Sequence[int]]:
        '''The size of each action fluent in canonical order.
        Returns:
            Sequence[Sequence[int]]: A tuple of tuple of integers
            representing the shape and size of each fluent.
        '''
        fluents = self.domain.action_fluents
        ordering = self.domain.action_fluent_ordering
        return self._fluent_size(fluents, ordering)

    @property
    def interm_size(self)-> Sequence[Sequence[int]]:
        '''The size of each intermediate fluent in canonical order.
        Returns:
            Sequence[Sequence[int]]: A tuple of tuple of integers
            representing the shape and size of each fluent.
        '''
        fluents = self.domain.intermediate_fluents
        ordering = self.domain.interm_fluent_ordering
        return self._fluent_size(fluents, ordering)

    @property
    def state_range_type(self) -> Sequence[str]:
        '''The range type of each state fluent in canonical order.
        Returns:
            Sequence[str]: A tuple of range types representing
            the range of each fluent.
        '''
        fluents = self.domain.state_fluents
        ordering = self.domain.state_fluent_ordering
        return self._fluent_range_type(fluents, ordering)

    @property
    def action_range_type(self) -> Sequence[str]:
        '''The range type of each action fluent in canonical order.
        Returns:
            Sequence[str]: A tuple of range types representing
            the range of each fluent.
        '''
        fluents = self.domain.action_fluents
        ordering = self.domain.action_fluent_ordering
        return self._fluent_range_type(fluents, ordering)

    @property
    def interm_range_type(self) -> Sequence[str]:
        '''The range type of each intermediate fluent in canonical order.
        Returns:
            Sequence[str]: A tuple of range types representing
            the range of each fluent.
        '''
        fluents = self.domain.intermediate_fluents
        ordering = self.domain.interm_fluent_ordering
        return self._fluent_range_type(fluents, ordering)

    @classmethod
    def _fluent_range_type(cls, fluents, ordering) -> Sequence[str]:
        '''Returns the range types of `fluents` following the given `ordering`.
        Returns:
            Sequence[str]: A tuple of range types representing
            the range of each fluent.
        '''
        range_types = []
        for name in ordering:
            fluent = fluents[name]
            range_type = fluent.range
            range_types.append(range_type)
        return tuple(range_types)

    def _fluent_params(self, fluents, ordering) -> FluentParamsList:
        '''Returns the instantiated `fluents` for the given `ordering`.
        For each fluent in `fluents`, it instantiates each parameter
        type w.r.t. the contents of the object table.
        Returns:
            Sequence[Tuple[str, List[str]]]: A tuple of pairs of fluent name
            and a list of instantiated fluents represented as strings.
        '''
        variables = []
        for fluent_id in ordering:
            fluent = fluents[fluent_id]
            param_types = fluent.param_types
            objects = ()
            names = []
            if param_types is None:
                names = [fluent.name]
            else:
                objects = tuple(self.object_table[ptype]['objects'] for ptype in param_types)
                for values in itertools.product(*objects):
                    values = ','.join(values)
                    var_name = '{}({})'.format(fluent.name, values)
                    names.append(var_name)
            variables.append((fluent_id, names))
        return tuple(variables)

    def _fluent_size(self, fluents, ordering) -> Sequence[Sequence[int]]:
        '''Returns the sizes of `fluents` following the given `ordering`.
        Returns:
            Sequence[Sequence[int]]: A tuple of tuple of integers
            representing the shape and size of each fluent.
        '''
        shapes = []
        for name in ordering:
            fluent = fluents[name]
            shape = self._param_types_to_shape(fluent.param_types)
            shapes.append(shape)
        return tuple(shapes)

    def _param_types_to_shape(self, param_types: Optional[str]) -> Sequence[int]:
        '''Returns the fluent shape given its `param_types`.'''
        param_types = [] if param_types is None else param_types
        shape = tuple(self.object_table[ptype]['size'] for ptype in param_types)
        return shape

    def get_dependencies(self, expr):
        deps = set()

        expressions = [expr]
        while expressions:
            expr = expressions.pop()

            for name in expr.scope:
                fluent, _ = self.fluent_table[name]

                if fluent.is_intermediate_fluent():
                    cpf = self.domain.get_intermediate_cpf(name)
                    expressions.append(cpf.expr)
                else:
                    deps.add(fluent)

        return deps



import importlib
import os

from pyRDDLGym.core.debug.exception import (
    RDDLEnvironmentNotExistError,
    RDDLInstanceNotExistError
)
from pyRDDLGym.core.env import RDDLEnv

VALID_EXT = '.rddl'
REPO_MANAGER_MODULE = 'rddlrepository'  # uses alias
REPO_MANAGER_CLASS = 'RDDLRepoManager'


def make(domain: str, instance: str, base_class=RDDLEnv, **env_kwargs) -> RDDLEnv:
    '''Creates a new RDDLEnv gym environment from domain and instance identifier
    or local file paths. 
    
    If the domain and instance arguments are paths to rddl files, then this 
    creates a RDDLEnv from the rddl files. If the domain and instance 
    arguments are general strings, then creates a RDDLEnv from rddlrepository.
    Note, that the second option requires rddlrepository to be installed and
    on the Python path, i.e. pip install rddlrepository.
    
    :param domain: the domain identifier, or path to domain rddl
    :param instance: the instance identifier, or path to instance rddl
    :param base_class: a subclass of RDDLEnv to load
    :param **env_kwargs: other arguments to pass to the RDDLEnv. 
    '''
    
    # check if arguments are file paths
    domain_is_file = os.path.isfile(domain)
    instance_is_file = os.path.isfile(instance)
    if domain_is_file != instance_is_file:
        raise ValueError(f'Domain and instance must either be both valid file '
                         f'paths or neither, but got {domain} and {instance}.')
    
    # if they are files, check they are RDDL files
    if domain_is_file:
        domain_is_rddl = str(os.path.splitext(domain)[1]).lower() == VALID_EXT
        instance_is_rddl = str(os.path.splitext(instance)[1]).lower() == VALID_EXT
        if not domain_is_rddl or not instance_is_rddl:
            raise ValueError(f'Domain and instance paths {domain} and {instance} '
                             f'are not valid RDDL ({VALID_EXT}) files.')
        
        # extract environment
        env = base_class(domain=domain, instance=instance, **env_kwargs)
        return env
    
    # check the repository exists
    spec = importlib.util.find_spec(REPO_MANAGER_MODULE)
    if spec is None:
        raise ImportError(f'rddlrepository is not installed: '
                          f'can be installed with \'pip install rddlrepository\'.')
    
    # load the repository manager
    module = importlib.import_module(REPO_MANAGER_MODULE)
    manager = getattr(module, REPO_MANAGER_CLASS)()
    info = manager.get_problem(domain)
        
    # extract environment
    domain_path = info.get_domain()
    instance_path = info.get_instance(instance)
    env = base_class(domain=domain_path, instance=instance_path, **env_kwargs)
    viz = info.get_visualizer()
    if viz is not None:
        env.set_visualizer(viz)
    return env



from abc import ABCMeta, abstractmethod
import numpy as np
import random
import gymnasium as gym
from typing import Dict
import sys

from pyRDDLGym.core.env import RDDLEnv
from pyRDDLGym.core.debug.exception import RDDLRandPolicyVecNotImplemented

class BaseAgent(metaclass=ABCMeta):
    '''Base class for policies.'''
    use_tensor_obs = False  # uses internal tensor representation of state
    
    @abstractmethod
    def sample_action(self, state: object) -> object:
        '''Samples an action from the current policy evaluated at the given state.
        
        :param state: the current state
        '''
        pass
    
    def reset(self) -> None:
        '''Resets the policy and prepares it for the next episode.'''
        pass
    
    def evaluate(self, env: RDDLEnv, episodes: int=1, 
                 verbose: bool=False, render: bool=False, 
                 seed: int=None) -> Dict[str, float]:
        '''Evaluates the current agent on the specified environment by simulating
        roll-outs. Returns a dictionary of summary statistics of the returns
        accumulated on the roll-outs.
        
        :param env: the environment
        :param episodes: how many episodes (trials) to perform
        :param verbose: whether to print the transition information to console
        at each step of the simulation
        :param render: visualize the domain using the env internal visualizer
        :param seed: optional RNG seed for the environment
        '''
        
        # check compatibility with environment
        if env.vectorized != self.use_tensor_obs:
            raise Exception(f'RDDLEnv vectorized flag must match use_tensor_obs '
                            f'of current policy, got {env.vectorized} and '
                            f'{self.use_tensor_obs}, respectively.')
        
        gamma = env.discount
        
        history = np.zeros((episodes,))
        for episode in range(episodes):
            
            # restart episode
            total_reward, cuml_gamma = 0.0, 1.0
            self.reset()
            state, _ = env.reset(seed=seed)
            
            # simulate to end of horizon
            for step in range(env.horizon):
                if render:
                    env.render()
                
                # take a step in the environment
                action = self.sample_action(state)   
                next_state, reward, terminated, truncated, _ = env.step(action)
                total_reward += reward * cuml_gamma
                cuml_gamma *= gamma
                done = terminated or truncated
                
                # printing
                if verbose: 
                    print(f'step       = {step}\n'
                          f'state      = {state}\n'
                          f'action     = {action}\n'
                          f'next state = {next_state}\n'
                          f'reward     = {reward}\n')
                state = next_state
                if done:
                    break
            
            if verbose:
                print(f'episode {episode + 1} ended with return {total_reward}')
            history[episode] = total_reward
        
        # summary statistics
        return {
            'mean': np.mean(history),
            'median': np.median(history),
            'min': np.min(history),
            'max': np.max(history),
            'std': np.std(history)
        }


class RandomAgent(BaseAgent):
    '''Uniformly pseudo-random policy.'''

    def __init__(self, action_space, num_actions=1, seed=None, noop_values=None):
        '''Creates a new uniformly pseudo-random policy.
        
        :param action_space: the set of actions from which to sample uniformly
        :param num_actions: the number of samples to produce
        '''
        self.action_space = action_space
        self.noop_values = noop_values
        self.action_buckets = []
        self.full_action_space = True

        self.vectorized = False
        for key, value in self.action_space.items():
            if len(value.shape) >= 1:
                if value.shape[0] > 1:
                    self.vectorized = True
                self.action_buckets.append(value.shape[0])
            else:
                self.action_buckets.append(1)
        if sum(self.action_buckets) - num_actions > 0:
            self.full_action_space = False

        if self.vectorized and not self.full_action_space:
            raise RDDLRandPolicyVecNotImplemented(
                    f'Random Agent does not support vectorized spaces for partial action specification, '
                    f'number of required actions is {num_actions}, where {sum(self.action_buckets)} is expected.')

        self.num_actions = num_actions
        self.rng = random.Random(seed)
        if seed is not None:
            self.action_space.seed(seed)

    def sample_action(self, state=None):
        s = self.action_space.sample()
        action = {}
        if not self.vectorized:
            selected_actions = self.rng.sample(list(s), self.num_actions)
            for sample in selected_actions:
                if isinstance(self.action_space[sample], gym.spaces.Box):
                    action[sample] = s[sample][0].item()
                elif isinstance(self.action_space[sample], gym.spaces.Discrete):
                    action[sample] = s[sample]
        else:
            action = s
        return action


class NoOpAgent(BaseAgent):
    '''No-op policy.'''
    
    def __init__(self, action_space, num_actions=0):
        '''Creates a new no-op policy.
            
        :param action_space: the set of actions (currently unused)
        :param num_actions: the number of samples to produce (currently unused)
        '''
        self.action_space = action_space
        self.num_actions = num_actions

    def sample_action(self, state=None):
        action = {}
        return action



import numpy as np
from typing import Dict, Union

from pyRDDLGym.core.compiler.model import RDDLPlanningModel
from pyRDDLGym.core.debug.exception import (
    RDDLInvalidObjectError,
    RDDLTypeError
)
from pyRDDLGym.core.debug.logger import Logger


class RDDLValueInitializer:
    '''Compiles all initial values in pvariables scope and init-fluents scope
    in a RDDL domain + instance to scalars or numpy arrays.
    '''
    
    INT = np.int64
    REAL = np.float64
        
    NUMPY_TYPES = {
        'int': INT,
        'real': REAL,
        'bool': bool
    }
    
    DEFAULT_VALUES = {
        'int': 0,
        'real': 0.0,
        'bool': False
    }
        
    def __init__(self, rddl: RDDLPlanningModel, logger: Logger=None) -> None:
        '''Creates a new object to compile initial values from a RDDL file. 
        Initial values of parameterized variables are stored in numpy arrays.
        For a variable var(?x1, ?x2, ... ?xn), the numpy array has n dimensions, 
        where the i-th dimension has number of elements equal to the number of 
        elements in the type of ?xi.
        
        :param rddl: the RDDL file whose initial values to extract
        :param logger: to log information about initial values to file
        '''
        self.rddl = rddl
        self.logger = logger
    
    def initialize(self) -> Dict[str, Union[np.ndarray, INT, REAL, bool]]:
        '''Compiles all initial values of all variables for the current RDDL file.
        A dictionary is returned with variable names as keys (as they appear in
        the RDDL) and value arrays as values.'''
        rddl = self.rddl
                
        # initial values consists of non-fluents, state and action fluents
        init_values = {}
        init_values.update(rddl.non_fluents)
        init_values.update(rddl.state_fluents)
        init_values.update(rddl.action_fluents)

        # domain objects are converted to integers
        for (var, values) in init_values.items():
            prange = rddl.variable_ranges[var]
            if prange in rddl.enum_types:
                init_values[var] = self._objects_to_ints(values, prange, var)
        
        # create a tensor for each pvar with the init_values
        # if the init_values are missing use the default value of range
        np_init_values = {}
        for (var, prange) in rddl.variable_ranges.items():
            
            # domain objects are treated as int
            if prange in rddl.enum_types:
                prange = 'int'
            
            # get default value and dtype
            default = RDDLValueInitializer.DEFAULT_VALUES.get(prange, None)
            dtype = RDDLValueInitializer.NUMPY_TYPES.get(prange, None)
            if default is None or dtype is None:
                raise RDDLTypeError(
                    f'Type <{prange}> of variable <{var}> is not valid, '
                    f'must be either an enumerated type in '
                    f'{rddl.enum_types} or an object type in '
                    f'{set(RDDLValueInitializer.DEFAULT_VALUES.keys())}.')
                        
            # convert a parameterized variable to dimensioned numpy array
            ptypes = rddl.variable_params[var]
            if ptypes:
                shape = rddl.object_counts(ptypes)     
                values = init_values.get(var, None)           
                if values is None:
                    values = np.full(shape=shape, fill_value=default, dtype=dtype)
                else:
                    values = np.reshape(
                        [(default if v is None else v) for v in values], 
                        newshape=shape, order='C')
                    
                    # cast to the required type
                    if not np.can_cast(values, dtype):
                        raise RDDLTypeError(
                            f'Initial values {values} for variable <{var}> '
                            f'cannot all be cast to required type <{prange}>.')
                    values = np.asarray(values, dtype=dtype)
            
            # convert scalar variable to scalar numpy array
            else:
                values = init_values.get(var, default)
                if isinstance(values, str) or not np.can_cast(values, dtype):
                    raise RDDLTypeError(
                        f'Initial value {values} for variable <{var}> '
                        f'cannot be cast to required type <{prange}>.')
                values = dtype(values)         
                       
            np_init_values[var] = values
        
        # log shapes of initial values
        if self.logger is not None:
            tensor_info = '\n\t'.join((
                f'{k}{rddl.variable_params[k]}, '
                f'shape={v.shape if type(v) is np.ndarray else ()}, '
                f'dtype={v.dtype if type(v) is np.ndarray else type(v).__name__}'
            ) for (k, v) in np_init_values.items())
            message = (
                f'[info] initializing pvariable tensors:' 
                f'\n\t{tensor_info}\n'
            )
            self.logger.log(message)
        
        return np_init_values
    
    def _objects_to_ints(self, literals, prange, var):
        is_scalar = isinstance(literals, str)
        if is_scalar:
            literals = [literals]
            
        indices = [0] * len(literals)
        for (i, obj) in enumerate(literals):
            if obj is not None:
                if self.rddl.object_to_type.get(obj, None) != prange:
                    raise RDDLInvalidObjectError(
                        f'<{obj}> assigned to pvariable <{var}> in instance '
                        f'is not an object of type <{prange}>.')
                indices[i] = self.rddl.object_to_index[obj]
                
        if is_scalar:
            indices, = indices
        return indices
    



import sys

import pyRDDLGym
from pyRDDLGym.core.server import RDDLSimServer


def main(domain, instance):
    env = pyRDDLGym.make(domain, instance)    
    agent = RDDLSimServer(env.domain_text, env.instance_text, 30, 300)
    agent.run()


if __name__ == "__main__":
    args = sys.argv[1:]
    domain, instance = "Wildfire", 0
    if len(args) == 2:
        domain, instance = args
    main(domain, instance)



import abc
import time

MAX_INT_32 = 2 ** 31 - 1


class RDDLEnvSeeder(metaclass=abc.ABCMeta):
    pass


class RDDLEnvSeederFibonacci(RDDLEnvSeeder):
    
    def __iter__(self):
        self.a = 1
        self.b = 2
        return self
    
    def __next__(self) -> int:
        v = self.a
        self.a, self.b = self.b, self.a + self.b
        if self.b > MAX_INT_32:
            self.a = self.a % MAX_INT_32
            self.b = self.b % MAX_INT_32
        return v


class RDDLEnvSeederTimestamp(RDDLEnvSeeder):
    
    def __iter__(self):
        return self
    
    def __next__(self) -> int:
        time_millis = int(time.time_ns())
        return time_millis % MAX_INT_32



import numpy as np

from pyRDDLGym.core.debug.exception import print_stack_trace, raise_warning
from pyRDDLGym.core.simulator import RDDLSimulator


class RDDLConstraints:
    '''Provides added functionality to understand a set of action-preconditions
    and state invariants in a RDDL file.
    '''
    
    def __init__(self, simulator: RDDLSimulator,
                 max_bound: float=np.inf,
                 inequality_tol: float=1e-5,
                 vectorized: bool=False) -> None:
        '''Creates a new set of state and action constraints.
        
        :param simulator: the RDDL simulator to evaluate potential non-fluent
        expressions nested in constraints
        :param max_bound: initial value for maximum possible bounds
        :param inequality_tol: tolerance for inequality > and < comparisons
        :param vectorized: whether bounds are represented as pairs of numpy arrays
        corresponding to lifted fluent names (if True), or as pairs of scalars for 
        grounded fluent names (if False)
        '''        
        self.sim = simulator
        self.rddl = rddl = simulator.rddl
        self.BigM = max_bound
        self.epsilon = inequality_tol
        self.vectorized = vectorized
        
        # initialize the bounds to [-inf, inf]
        self._bounds = {}
        for (var, vtype) in rddl.variable_types.items():
            if vtype in {'state-fluent', 'observ-fluent', 'action-fluent'}:
                ptypes = rddl.variable_params[var]
                shape = rddl.object_counts(ptypes)
                if shape:
                    self._bounds[var] = [
                        np.full(shape=shape, fill_value=-self.BigM),
                        np.full(shape=shape, fill_value=+self.BigM)
                    ]
                else:
                    self._bounds[var] = [-self.BigM, +self.BigM]

        # actions and states bounds extraction for gym's action and state spaces
        # currently supports only linear inequality constraints
        self._is_box_precond = []
        for (index, precond) in enumerate(rddl.preconditions):
            tag = f'Action precondition {index + 1}'
            is_box = self._parse_bounds(tag, precond, [], rddl.action_fluents)
            self._is_box_precond.append(is_box)
                    
        self._is_box_invariant = []
        for (index, invariant) in enumerate(rddl.invariants):
            tag = f'State invariant {index + 1}'
            is_box = self._parse_bounds(tag, invariant, [], rddl.state_fluents)
            self._is_box_invariant.append(is_box)

        for (name, bounds) in self._bounds.items():
            RDDLSimulator._check_bounds(*bounds, f'Variable <{name}>', bounds)
        
        # ground the bounds if not vectorized
        if self.vectorized:
            self._bounds = {name: tuple(value) 
                            for (name, value) in self._bounds.items()}
        else:
            new_bounds = {}
            for (var, (lower, upper)) in self._bounds.items():
                lower = np.ravel(lower, order='C')
                upper = np.ravel(upper, order='C')
                gvars = rddl.variable_groundings[var]
                assert (len(gvars) == len(lower) == len(upper))
                new_bounds.update(zip(gvars, zip(lower, upper)))      
            self._bounds = new_bounds        
        
        # log bounds to file
        if simulator.logger is not None:
            bounds_info = '\n\t'.join(f'{k}: {v}' 
                                      for (k, v) in self._bounds.items())
            simulator.logger.log(f'[info] computed simulation bounds:\n' 
                                 f'\t{bounds_info}\n')
        
    def _parse_bounds(self, tag, expr, objects, search_vars):
        etype, op = expr.etype
        
        # for aggregation can only parse forall, since it is equivalent to a 
        # set of independent constraints determined by the function aggregated
        if etype == 'aggregation' and op == 'forall':
            * pvars, arg = expr.args
            new_objects = objects + [pvar for (_, pvar) in pvars]
            return self._parse_bounds(tag, arg, new_objects, search_vars)
        
        # for logical expression can only parse conjunction
        # same rationale as forall discussed above
        elif etype == 'boolean' and op == '^':
            success = True
            for arg in expr.args:
                success_arg = self._parse_bounds(tag, arg, objects, search_vars)
                success = success and success_arg
            return success
        
        # relational operation in constraint at the top level, i.e. constraint
        # LHS <= RHS for example
        elif etype == 'relational':
            rddl = self.rddl
            var, lim, loc, slices = self._parse_bounds_relational(
                tag, expr, objects, search_vars)
            success = var is not None and loc is not None
            if success: 
                op = np.minimum if loc == 1 else np.maximum
                if slices:
                    self._bounds[var][loc][slices] = op(
                        self._bounds[var][loc][slices], lim)
                else:
                    self._bounds[var][loc] = op(self._bounds[var][loc], lim)
            return success
        
        # not possible to parse as a box constraint
        else:
            return False
               
    def _parse_bounds_relational(self, tag, expr, objects, search_vars):
        left, right = expr.args    
        _, op = expr.etype
        left_pvar = left.is_pvariable_expression() and left.args[0] in search_vars
        right_pvar = right.is_pvariable_expression() and right.args[0] in search_vars
        
        # both LHS and RHS are pvariable expressions, or relational operator 
        # cannot be simplified further
        if (left_pvar and right_pvar) or op not in {'<=', '<', '>=', '>'}:
            raise_warning(
                f'{tag} does not have a structure of '
                f'<action or state fluent> <op> <rhs>, where ' 
                f'<op> is one of {{<=, <, >=, >}} and '
                f'<rhs> is a deterministic function of non-fluents only, '
                f'and will be ignored.\n' + 
                print_stack_trace(expr), 'red')
            return None, 0.0, None, []
        
        # neither side is a pvariable, nothing to do
        elif not left_pvar and not right_pvar:
            return None, 0.0, None, []
        
        # only one of LHS and RHS are pvariable expressions, other side constant
        else:
            
            # which one is the constant expression?
            if left_pvar:
                var, args = left.args
                const_expr = right
            else:
                var, args = right.args
                const_expr = left
            if args is None:
                args = []
                
            if not self.rddl.is_non_fluent_expression(const_expr):
                raise_warning(
                    f'{tag} contains a fluent expression '
                    f'(a nondeterministic operation or fluent variable) '
                    f'on both sides of an (in)equality, and will be ignored.\n' + 
                    print_stack_trace(const_expr), 'red')
                return None, 0.0, None, []
            
            # use the simulator to evaluate the constant side of the comparison
            # this is likened to a constant-folding operation
            const = self.sim._sample(const_expr, self.sim.subs)
            eps, loc = self._get_op_code(op, left_pvar)
            lim = const + eps
            
            # finally, since the pvariable expression may contain literals,
            # construct a slice that would assign the above constant value to it
            rddl = self.rddl
            slices = []
            for arg in args:
                if rddl.is_literal(arg):
                    arg = rddl.strip_literal(arg)
                    slices.append(rddl.object_to_index[arg])
                else:
                    slices.append(slice(None))
            slices = tuple(slices)
            
            return var, lim, loc, slices
            
    def _get_op_code(self, op, is_right):
        eps = 0.0
        if is_right:
            if op in ['<=', '<']:
                loc = 1
                if op == '<':
                    eps = -self.epsilon
            elif op in ['>=', '>']:
                loc = 0
                if op == '>':
                    eps = self.epsilon
        else:
            if op in ['<=', '<']:
                loc = 0
                if op == '<':
                    eps = self.epsilon
            elif op in ['>=', '>']:
                loc = 1
                if op == '>':
                    eps = -self.epsilon
        return eps, loc

    @property
    def bounds(self):
        return self._bounds

    @bounds.setter
    def bounds(self, value):
        self._bounds = value
    
    @property
    def is_box_preconditions(self):
        return self._is_box_precond
    
    @property
    def is_box_invariants(self):
        return self._is_box_invariant
    



# This file is based on thiago pbueno's pyrddl.
# https://github.com/thiagopbueno/pyrddl
# it was adapted and extended for pyRDDLGym

from typing import Dict, List, Sequence, Optional, Tuple

from pyRDDLGym.core.parser import utils
from pyRDDLGym.core.parser.pvariable import PVariable
from pyRDDLGym.core.parser.cpf import CPF
from pyRDDLGym.core.parser.expr import Expression

Type = Tuple[str, str]


class Domain(object):
    '''Domain class for accessing RDDL domain sections.
    Note:
        This class is intended to be solely used by the parser and compiler.
        Do not attempt to directly use this class to build a Domain object.
    Args:
        name: Name of RDDL domain.
        requirements: List of RDDL requirements.
        sections: Mapping from string to domain section.
    Attributes:
        name (str): Domain identifier.
        requirements (List[str]): List of requirements.
        types (List[:obj:`Type`]): List of types.
        pvariables (List[:obj:`PVariable`]): List of parameterized variables.
        cpfs (List[:obj:`CPF`]): List of Conditional Probability Functions.
        reward (:obj:`Expression`): Reward function.
        preconds (List[:obj:`Expression`]): List of action preconditions.
        terminals (List[:obj:`Expression`]): List of termination conditions.
        constraints (List[:obj:`Expression`]): List of state-action constraints.
        invariants (List[:obj:`Expression`]): List of state invariants.
    '''

    def __init__(self, name: str, requirements: List[str], sections: Dict[str, Sequence]) -> None:
        self.name = name
        self.requirements = requirements

        self.pvariables = sections['pvariables']
        self.cpfs = sections['cpfs']
        self.reward = sections['reward']

        self.types = sections.get('types', [])
        self.preconds = sections.get('preconds', [])
        self.terminals = sections.get('terminals', [])
        self.invariants = sections.get('invariants', [])
        self.constraints = sections.get('constraints', [])

    def build(self):
        self._build_preconditions_table()
        self._build_action_bound_constraints_table()

    def _build_preconditions_table(self):
        '''Builds the local action precondition expressions.'''
        self.local_action_preconditions = dict()
        self.global_action_preconditions = []
        action_fluents = self.action_fluents
        for precond in self.preconds:
            scope = precond.scope
            action_scope = [action for action in scope if action in action_fluents]
            if len(action_scope) == 1:
                name = action_scope[0]
                self.local_action_preconditions[name] = self.local_action_preconditions.get(name, [])
                self.local_action_preconditions[name].append(precond)
            else:
                self.global_action_preconditions.append(precond)

    def _build_action_bound_constraints_table(self):
        '''Builds the lower and upper action bound constraint expressions.'''
        self.action_lower_bound_constraints = {}
        self.action_upper_bound_constraints = {}

        for name, preconds in self.local_action_preconditions.items():

            for precond in preconds:
                expr_type = precond.etype
                expr_args = precond.args

                bounds_expr = None

                if expr_type == ('aggregation', 'forall'):
                    inner_expr = expr_args[1]
                    if inner_expr.etype[0] == 'relational':
                        bounds_expr = inner_expr
                elif expr_type[0] == 'relational':
                    bounds_expr = precond

                if bounds_expr:
                    # lower bound
                    bound = self._extract_lower_bound(name, bounds_expr)
                    if bound is not None:
                        self.action_lower_bound_constraints[name] = bound
                    else:  # upper bound
                        bound = self._extract_upper_bound(name, bounds_expr)
                        if bound is not None:
                            self.action_upper_bound_constraints[name] = bound

    def _extract_lower_bound(self, name: str, expr: Expression) -> Optional[Expression]:
        '''Returns the lower bound expression of the action with given `name`.'''
        etype = expr.etype
        args = expr.args
        if etype[1] in ['<=', '<']:
            if args[1].is_pvariable_expression() and args[1].name == name:
                return args[0]
        elif etype[1] in ['>=', '>']:
            if args[0].is_pvariable_expression() and args[0].name == name:
                return args[1]
        return None

    def _extract_upper_bound(self, name: str, expr: Expression) -> Optional[Expression]:
        '''Returns the upper bound expression of the action with given `name`.'''
        etype = expr.etype
        args = expr.args
        if etype[1] in ['<=', '<']:
            if args[0].is_pvariable_expression() and args[0].name == name:
                return args[1]
        elif etype[1] in ['>=', '>']:
            if args[1].is_pvariable_expression() and args[1].name == name:
                return args[0]
        return None

    @property
    def non_fluents(self) -> Dict[str, PVariable]:
        '''Returns non-fluent pvariables.'''
        return { str(pvar): pvar for pvar in self.pvariables if pvar.is_non_fluent() }

    @property
    def state_fluents(self) -> Dict[str, PVariable]:
        '''Returns state-fluent pvariables.'''
        return { str(pvar): pvar for pvar in self.pvariables if pvar.is_state_fluent() }

    @property
    def action_fluents(self) -> Dict[str, PVariable]:
        '''Returns action-fluent pvariables.'''
        return { str(pvar): pvar for pvar in self.pvariables if pvar.is_action_fluent() }

    @property
    def intermediate_fluents(self) -> Dict[str, PVariable]:
        '''Returns interm-fluent pvariables.'''
        return { str(pvar): pvar for pvar in self.pvariables if pvar.is_intermediate_fluent() }

    @property
    def derived_fluents(self) -> Dict[str, PVariable]:
        '''Returns derived-fluent pvariables.'''
        return {str(pvar): pvar for pvar in self.pvariables if pvar.is_derived_fluent()}

    @property
    def observation_fluents(self) -> Dict[str, PVariable]:
        '''Returns observ-fluents pvaraibles'''
        return {str(pvar): pvar for pvar in self.pvariables if pvar.is_observ_fluent()}
        return

    @property
    def intermediate_cpfs(self) -> List[CPF]:
        '''Returns list of intermediate-fluent CPFs in level order.'''
        _, cpfs = self.cpfs
        interm_cpfs = [cpf for cpf in cpfs if cpf.name in self.intermediate_fluents]
        interm_cpfs = sorted(interm_cpfs, key=lambda cpf: (self.intermediate_fluents[cpf.name].level, cpf.name))
        return interm_cpfs

    @property
    def derived_cpfs(self) -> List[CPF]:
        '''Returns list of derived-fluent CPFs in level order.'''
        _, cpfs = self.cpfs
        der_cpfs = [cpf for cpf in cpfs if cpf.name in self.derived_fluents]
        der_cpfs = sorted(der_cpfs, key=lambda cpf: (self.derived_fluents[cpf.name].level, cpf.name))
        return der_cpfs

    @property
    def observation_cpfs(self) -> List[CPF]:
        '''Returns list of observation_fluents CPFs in level order.'''
        _, cpfs = self.cpfs
        der_cpfs = [cpf for cpf in cpfs if cpf.name in self.observation_fluents]
        der_cpfs = sorted(der_cpfs, key=lambda cpf: (self.observation_fluents[cpf.name].level, cpf.name))
        return der_cpfs

    def get_intermediate_cpf(self, name):
        for cpf in self.intermediate_cpfs:
            if cpf.name == name:
                return cpf

    def get_derived_cpf(self, name):
        for cpf in self.derived_cpfs:
            if cpf.name == name:
                return cpf

    @property
    def state_cpfs(self) -> List[CPF]:
        '''Returns list of state-fluent CPFs.'''
        _, cpfs = self.cpfs
        state_cpfs = []
        for cpf in cpfs:
            name = utils.rename_next_state_fluent(cpf.name)
            if name in self.state_fluents:
                state_cpfs.append(cpf)
        state_cpfs = sorted(state_cpfs, key=lambda cpf: cpf.name)
        return state_cpfs

    @property
    def non_fluent_ordering(self) -> List[str]:
        '''The list of non-fluent names in canonical order.
        Returns:
            List[str]: A list of fluent names.
        '''
        return sorted(self.non_fluents)

    @property
    def state_fluent_ordering(self) -> List[str]:
        '''The list of state-fluent names in canonical order.
        Returns:
            List[str]: A list of fluent names.
        '''
        return sorted(self.state_fluents)

    @property
    def action_fluent_ordering(self) -> List[str]:
        '''The list of action-fluent names in canonical order.
        Returns:
            List[str]: A list of fluent names.
        '''
        return sorted(self.action_fluents)

    @property
    def interm_fluent_ordering(self) -> List[str]:
        '''The list of intermediate-fluent names in canonical order.
        Returns:
            List[str]: A list of fluent names.
        '''
        interm_fluents = self.intermediate_fluents.values()
        key = lambda pvar: (pvar.level, pvar.name)
        return [str(pvar) for pvar in sorted(interm_fluents, key=key)]

    @property
    def derived_fluent_ordering(self) -> List[str]:
        '''The list of derived-fluent names in canonical order.
        Returns:
            List[str]: A list of fluent names.
        '''
        derived_fluents = self.derived_fluents.values()
        key = lambda pvar: (pvar.level, pvar.name)
        return [str(pvar) for pvar in sorted(derived_fluents, key=key)]

    @property
    def next_state_fluent_ordering(self) -> List[str]:
        '''The list of next state-fluent names in canonical order.
        Returns:
            List[str]: A list of fluent names.
        '''
        key = lambda x: x.name
        return [cpf.name for cpf in sorted(self.state_cpfs, key=key)]



import gymnasium as gym
from gymnasium.spaces import Box, Dict, Discrete
import numpy as np
import os
import pygame
import typing

from pyRDDLGym.core.compiler.model import RDDLLiftedModel
from pyRDDLGym.core.constraints import RDDLConstraints
from pyRDDLGym.core.debug.exception import (
    RDDLEpisodeAlreadyEndedError,
    RDDLLogFolderError,
    RDDLTypeError
)
from pyRDDLGym.core.debug.logger import Logger, SimLogger
from pyRDDLGym.core.parser.parser import RDDLParser
from pyRDDLGym.core.parser.reader import RDDLReader
from pyRDDLGym.core.seeding import RDDLEnvSeeder, RDDLEnvSeederFibonacci
from pyRDDLGym.core.simulator import RDDLSimulator
from pyRDDLGym.core.visualizer.chart import ChartVisualizer
from pyRDDLGym.core.visualizer.heatmap import HeatmapVisualizer
from pyRDDLGym.core.visualizer.text import TextVisualizer


def _make_dir(log_path):
    root_path = os.path.dirname(log_path)
    if not os.path.exists(root_path):
        try:
            os.makedirs(root_path)
        except Exception as e:
            if not isinstance(e, FileExistsError):
                raise RDDLLogFolderError(
                    f'Could not create folder at path {root_path}.')
    return log_path

    
class RDDLEnv(gym.Env):
    '''A gym environment class for RDDL domains.'''
    
    def __init__(self, domain: str,
                 instance: str,
                 enforce_action_constraints: bool=False,
                 enforce_action_count_non_bool: bool=True,
                 vectorized: bool=False,
                 debug: bool=False,
                 log_path: str=None,
                 backend: RDDLSimulator=RDDLSimulator,
                 backend_kwargs: typing.Dict={},
                 seeds: RDDLEnvSeeder=RDDLEnvSeederFibonacci()):
        '''Creates a new gym environment from the given RDDL domain + instance.
        
        :param domain: the RDDL domain
        :param instance: the RDDL instance
        :param enforce_action_constraints: whether to raise an exception if the
        action constraints are violated
        :param enforce_action_count_non_bool: whether to include non-bool actions
        in check that number of nondef actions don't exceed max-nondef-actions
        :param vectorized: whether actions and states are represented as
        dictionaries of numpy arrays (if True), or as dictionaries of scalars
        :param debug: whether to log compilation information to a log file
        :param log_path: absolute path to file where simulation log is saved,
        excluding the file extension, None means no logging
        :param backend: the subclass of RDDLSimulator to use as backend for
        simulation (currently supports numpy and Jax)
        :param backend_kwargs: dictionary of additional named arguments to
        pass to backend (must not include logger)
        :param seeds: an instance of RDDLEnvSeeder for generating RNG seeds
        '''
        super(RDDLEnv, self).__init__()
        
        self.domain_text = domain
        self.instance_text = instance
        self.enforce_action_constraints = enforce_action_constraints
        self.enforce_count_non_bool = enforce_action_count_non_bool
        self.vectorized = vectorized
        
        # read and parse domain and instance
        reader = RDDLReader(domain, instance)
        domain = reader.rddltxt
        parser = RDDLParser(lexer=None, verbose=False)
        parser.build()
        rddl = parser.parse(domain)
        
        # define the RDDL model
        self.model = RDDLLiftedModel(rddl)
        self.horizon = self.model.horizon
        self.discount = self.model.discount
        self.max_allowed_actions = self.model.max_allowed_actions 
                
        # for logging compilation data
        log_fname = f'{self.model.domain_name}_{self.model.instance_name}'
        logger = Logger(f'{log_fname}_debug.log') if debug else None
        self.logger = logger
        
        # for logging simulation data
        self.simlogger = None
        if log_path is not None and log_path:
            new_log_path = _make_dir(log_path)
            self.simlogger = SimLogger(f'{new_log_path}_log.csv')
            self.simlogger.clear(overwrite=False)
        
        # define the simulation backend  
        self.sampler = backend(self.model,
                               logger=logger,
                               keep_tensors=self.vectorized,
                               **backend_kwargs)
        
        # compute the bounds on fluents from the constraints
        constraints = RDDLConstraints(self.sampler, vectorized=self.vectorized)
        self._bounds = constraints.bounds
        self._shapes = {var: np.shape(values[0]) 
                        for (var, values) in self._bounds.items()}
        
        # construct the gym observation space
        if self.sampler.is_pomdp:
            state_ranges = self.model.observ_ranges
        else:
            state_ranges = self.model.state_ranges
        if not self.vectorized:
            state_ranges = self.model.ground_vars_with_value(state_ranges)   
             
        self.observation_space = self._rddl_to_gym_bounds(state_ranges)
        
        # construct the gym action space      
        if self.vectorized:
            self._action_ranges = self.model.action_ranges
            self._noop_actions = self.sampler.noop_actions
        else:
            self._action_ranges = self.sampler.grounded_action_ranges
            self._noop_actions = self.sampler.grounded_noop_actions
            
        self.action_space = self._rddl_to_gym_bounds(self._action_ranges)
        
        # set the visualizer
        self._visualizer = ChartVisualizer(self.model)
        self._movie_generator = None
        self.state = None
        self.image = None
        self.window = None
        self.to_render = False
        self.image_size = None
        
        # set roll-out parameters           
        self.trial = 0
        self.timestep = 0
        self.done = False
        self.seeds = iter(seeds)
            
    def _rddl_to_gym_bounds(self, ranges):
        result = Dict()
        for (var, prange) in ranges.items():
            shape = self._shapes[var]
            
            # enumerated values converted to Discrete space
            if prange in self.model.enum_types:
                num_objects = len(self.model.type_to_objects[prange])
                if self.vectorized:
                    result[var] = Box(0, num_objects - 1, shape=shape, dtype=np.int32)
                else:
                    result[var] = Discrete(num_objects) 
            
            # real values define a box
            elif prange == 'real':
                low, high = self._bounds[var]
                result[var] = Box(low, high, dtype=np.float32)
            
            # boolean values converted to Discrete space
            elif prange == 'bool':
                if self.vectorized:
                    result[var] = Box(0, 1, shape=shape, dtype=np.int32)
                else:
                    result[var] = Discrete(2)
            
            # integer values converted to Discrete space
            elif prange == 'int':
                low, high = self._bounds[var]
                low = np.maximum(low, np.iinfo(np.int32).min)
                high = np.minimum(high, np.iinfo(np.int32).max)
                if self.vectorized:
                    result[var] = Box(low, high, shape=shape, dtype=np.int32)
                else:
                    result[var] = Discrete(int(high - low + 1), start=int(low))
            
            # unknown type
            else:
                raise RDDLTypeError(
                    f'Type <{prange}> of fluent <{var}> is not valid, '
                    f'must be an enumerated or primitive type (real, int, bool).')
                
        return result
    
    def seed(self, seed=None):
        self.sampler.seed(seed)
        return [seed]
    
    VISUALIZER_CLASSES = {
        'chart': ChartVisualizer,
        'heatmap': HeatmapVisualizer,
        'text': TextVisualizer
    }
    
    def set_visualizer(self, viz, movie_gen=None, movie_per_episode=False, **viz_kwargs):
        if viz is None:
            self._visualizer = ChartVisualizer(self.model)
        else:
            if isinstance(viz, str):
                viz = self.VISUALIZER_CLASSES.get(viz.lower(), None)
                if viz is None:
                    raise ValueError(
                        f'Visualizer type {viz} is invalid, '
                        f'must be one of {list(self.VISUALIZER_CLASSES.keys())}')
            self._visualizer = viz(self.model, **viz_kwargs)
        self._movie_generator = movie_gen
        self._movie_per_episode = movie_per_episode
        self._movies = 0
        self.to_render = False
    
    def _fix_boolean_actions(self, actions):
        fixed_actions = self._noop_actions.copy()
        for (var, values) in actions.items():
            if self._action_ranges.get(var, '') == 'bool':
                if np.shape(values):
                    fixed_actions[var] = np.asarray(values, dtype=bool)
                else:
                    fixed_actions[var] = bool(values)
            else:
                fixed_actions[var] = values
        return fixed_actions

    def step(self, actions):
        sampler = self.sampler
        
        if self.done:
            raise RDDLEpisodeAlreadyEndedError(
                'The step() function has been called even though the '
                'current episode has terminated or truncated: please call reset().')
            
        # fix actions and check constraints
        actions = self._fix_boolean_actions(actions)
        sampler.check_default_action_count(actions, self.enforce_count_non_bool)
        if self.enforce_action_constraints:
            sampler.check_action_preconditions(actions, silent=False)
        
        # sample next state and reward
        obs, reward, terminated = sampler.step(actions)
        self.state = sampler.states
            
        # produce array outputs for vectorized option
        if self.vectorized:
            obs = {var: np.atleast_1d(value) for (var, value) in obs.items()}
            
        # check if the state invariants are satisfied
        truncated = not sampler.check_state_invariants(silent=True)
        self.done = terminated or truncated
            
        # log to file
        if self.simlogger is not None:
            if self.vectorized:
                log_obs = self.model.ground_vars_with_values(obs)
                log_action = self.model.ground_vars_with_values(actions)
            else:
                log_obs = obs
                log_action = actions
            self.simlogger.log(log_obs, log_action, reward, self.done, self.timestep)
        
        # update step horizon
        self.timestep += 1
        if self.timestep == self.horizon:
            truncated = True
            self.done = True
        
        return obs, reward, terminated, truncated, {}

    def reset(self, seed=None, options=None):
        sampler = self.sampler
        
        # reset counters and internal state
        obs, terminated = sampler.reset()
        self.done = terminated
        self.state = sampler.states
        self.trial += 1
        self.timestep = 0
        
        # produce array outputs for vectorized option
        if self.vectorized:
            obs = {var: np.atleast_1d(value) for (var, value) in obs.items()}
            
        # update movie generator
        if self._movie_generator is not None and self._visualizer is not None:
            if self.vectorized:
                state = self.model.ground_vars_with_values(self.state)
            else:
                state = self.state
                
            image = self._visualizer.render(state)
            self.image_size = image.size
            if self._movie_per_episode:
                self._movie_generator.save_animation(
                    self._movie_generator.env_name + '_' + str(self._movies))
                self._movies += 1
            self._movie_generator.save_frame(image)            
        
        # update random generator seed
        if seed is not None:
            self.seed(seed)
        else:
            seed = next(self.seeds)
            if seed is not None:
                self.seed(seed)

        # logging
        if self.simlogger:
            text = (f'######################################################\n'
                    f'New Trial, seed={seed}\n'
                    f'######################################################')
            self.simlogger.log_free(text)
            
        return obs, {}

    def pilImageToSurface(self, pilImage):
        return pygame.image.fromstring(
            pilImage.tobytes(), pilImage.size, pilImage.mode).convert()

    def render(self, to_display=True):
        image = None
        if self._visualizer is not None:
            if self.vectorized:
                state = self.model.ground_vars_with_values(self.state)
            else:
                state = self.state
            
            # update the screen
            image = self._visualizer.render(state)
            self.image_size = image.size
            if to_display:
                if not self.to_render:
                    self.to_render = True
                    pygame.init()
                    self.window = pygame.display.set_mode(
                        (self.image_size[0], self.image_size[1]))
                self.window.fill(0)
                pygameSurface = self.pilImageToSurface(image)
                self.window.blit(pygameSurface, (0, 0))
                pygame.display.set_caption(self.model.instance_name)
                pygame.display.flip()
                
                # prevents the window from freezing up midway
                for event in pygame.event.get():
                    if event.type == pygame.QUIT:
                        pass
            
            # capture frame to disk
            if self._movie_generator is not None:
                self._movie_generator.save_frame(image)
    
        return image
    
    def close(self):
        if self.simlogger:
            self.simlogger.close()
        
        # close rendering and save animation  
        if self.to_render:
            pygame.display.quit()
            pygame.quit()    
            
            # prepare the animation from captured frames
            if self._movie_generator is not None:
                self._movie_generator.save_animation(
                    self._movie_generator.env_name + '_' + str(self._movies))
                self._movies += 1
